import warnings
from numbers import Number

from phi import math, struct, field
from phi.field import CenteredGrid, StaggeredGrid, GeometryMask, PointCloud, Field, HardGeometryMask, Grid
from phi.geom import Box, GridCell, Sphere, union, assert_same_rank
from phi.geom import Geometry
from phi.math import extrapolation, Tensor
from phi.math import spatial_shape
from ._effect import FieldEffect
from ._physics import Physics
from ._physics import State
from ..math.extrapolation import combine_sides


class Material:
    """
    Defines the extrapolation modes / boundary conditions for a surface.
    The surface can be an Obstacle or a Domain boundary.
    
    Use Material.as_material() to mix different materials for different sides.
    """

    def __init__(self, name: str, scalar_extrapolation, vector_extrapolation, near_vector_extrapolation, active_extrapolation, accessible_extrapolation):
        """
        Create a Material for a Domain or Obstacle.

        Args:
          name: material name
          scalar_extrapolation: extrapolation mode of grids created via Domain.grid()
          vector_extrapolation: extrapolation mode of grids created via Domain.vector_grid() or Domain.staggered_grid()
          near_vector_extrapolation: Used in pressure solve.
          active_extrapolation: Whether cells outside the domain bounds also belong to the domain. Used in pressure solve.
          accessible_extrapolation: Whether quantities can move in and out of the domain. Used in pressure solve.
        """
        self.name = name
        """ Material name """
        self.scalar_extrapolation = scalar_extrapolation
        """ Extrapolation mode of grids created via Domain.scalar_grid() """
        self.vector_extrapolation = vector_extrapolation
        """ Extrapolation mode of grids created via Domain.vector_grid() or Domain.staggered_grid() """
        self.near_vector_extrapolation = near_vector_extrapolation
        """ Used in pressure solve. """
        self.active_extrapolation = active_extrapolation
        """ Whether cells outside the domain bounds also belong to the domain. Used in pressure solve. """
        self.accessible_extrapolation = accessible_extrapolation
        """ Whether quantities can move in and out of the domain. Used in pressure solve. """

    def __repr__(self):
        return self.name

    @staticmethod
    def as_material(obj: 'Material' or tuple or list or dict) -> 'Material':
        """
        Construct a mixed material from from a sequence of materials.

        Args:
          obj: sequence of materials
          obj: Material or tuple or list or dict: 

        Returns:
          : Single mixed Material

        """
        if isinstance(obj, Material):
            return obj
        if isinstance(obj, (tuple, list)):
            dims = [math.GLOBAL_AXIS_ORDER.axis_name(i, len(obj)) for i in range(len(obj))]
            obj = {dim: mat for dim, mat in zip(dims, obj)}
        if isinstance(obj, dict):
            grid_extrapolation = _mix(obj, 'scalar_extrapolation')
            near_vector_extrapolation = _mix(obj, 'near_vector_extrapolation')
            vector_extrapolation = _mix(obj, 'vector_extrapolation')
            active_extrapolation = _mix(obj, 'active_extrapolation')
            accessible_extrapolation = _mix(obj, 'accessible_extrapolation')
            return Material('mixed', grid_extrapolation, near_vector_extrapolation, vector_extrapolation, active_extrapolation, accessible_extrapolation)
        raise NotImplementedError()


def _mix(material_dict, ext_property: str):
    extrapolations = {}
    for dim, material in material_dict.items():
        if isinstance(material, Material):
            extrapolations[dim] = getattr(material, ext_property)
        else:
            extrapolations[dim] = (getattr(material[0], ext_property), getattr(material[1], ext_property))
    return combine_sides(extrapolations)


OPEN = Material('open', extrapolation.ZERO, extrapolation.ZERO, extrapolation.BOUNDARY, extrapolation.ZERO, extrapolation.ONE)
CLOSED = NO_STICK = SLIPPERY = Material('slippery', extrapolation.BOUNDARY, extrapolation.BOUNDARY, extrapolation.ZERO, extrapolation.ZERO, extrapolation.ZERO)
NO_SLIP = STICKY = Material('sticky', extrapolation.BOUNDARY, extrapolation.ZERO, extrapolation.ZERO, extrapolation.ZERO, extrapolation.ZERO)
PERIODIC = Material('periodic', extrapolation.PERIODIC, extrapolation.PERIODIC, extrapolation.PERIODIC, extrapolation.ONE, extrapolation.ONE)


class Domain:

    def __init__(self, resolution: math.Shape = math.EMPTY_SHAPE, boundaries: Material or tuple or list = OPEN, bounds: Box = None, **resolution_):
        """
        The Domain specifies the grid resolution, physical size and boundary conditions of a simulation.

        It provides convenience methods for creating Grids fitting the domain, e.g. `grid()`, `vector_grid()` and `staggered_grid()`.

        Also see the `phi.physics` module documentation at https://tum-pbs.github.io/PhiFlow/Physics.html

        Args:
          resolution: grid dimensions as Shape or sequence of integers. Alternatively, dimensions can be specified directly as kwargs.
          boundaries: specifies the extrapolation modes of grids created from this Domain as a Material instance.
            Default materials include OPEN, CLOSED, PERIODIC.
            To specify boundary conditions per face of the domain, pass a sequence of Materials or Material pairs (lower, upper)., e.g. [CLOSED, (CLOSED, OPEN)].
          bounds: physical size of the domain. If not provided, the size is equal to the resolution (unit cubes).
        """
        self.resolution = spatial_shape(resolution) & spatial_shape(resolution_)
        """ Grid dimensions as `Shape` object containing spatial dimensions only. """
        self.boundaries = Material.as_material(boundaries)
        """ Outer boundary condition as `Material` object """
        self.bounds = Box(0, math.tensor(self.resolution, names='vector')) if bounds is None else bounds
        """ Physical dimensions of the domain as `Box` object """

    def __repr__(self):
        return '(%s, size=%s)' % (self.resolution, self.bounds.size)

    @property
    def shape(self):
        """ Alias for `Domain.resolution` """
        return self.resolution

    @property
    def rank(self):
        """Number of spatial dimensions of the simulation; spatial rank. 1 = 1D, 2 = 2D, 3 = 3D, etc."""
        return self.resolution.rank

    @property
    def dx(self):
        """Size of a single grid cell (physical size divided by resolution) as `Tensor`"""
        return self.bounds.size / self.resolution

    @property
    def cells(self):
        """
        Returns the geometry of all cells as a `Box` object.
        The box will have spatial dimensions matching the resolution of the Domain, i.e. `domain.cells.shape == domain.resolution`.
        """
        return GridCell(self.resolution, self.bounds)

    def center_points(self):
        """
        Returns a Tensor enumerating the physical center locations of all cells within the Domain.
        This is equivalent to calling `domain.cells.center`.
        
        The shape of the returned Tensor extends the domain resolution by one vector dimension.
        """
        return self.cells.center

    def grid(self,
             value: Field or Tensor or Number or Geometry or callable = 0,
             type: type = CenteredGrid,
             extrapolation: math.Extrapolation = None) -> CenteredGrid or StaggeredGrid:
        """
        *Deprecated* due to inconsistent extrapolation selection. Use `scalar_grid()` or `vector_grid()` instead.

        Creates a grid matching the resolution and bounds of the domain.
        The grid is created from the given `value` which must be one of the following:
        
        * Number (int, float, complex or zero-dimensional tensor): all grid values will be equal to `value`. This has a near-zero memory footprint.
        * Field: the given value is resampled to the grid cells of this Domain.
        * Tensor with spatial dimensions matching the domain resolution: grid values will equal `value`.
        * Geometry: grid values are determined from the volume overlap between grid cells and geometry. Non-overlapping = 0, fully enclosed grid cell = 1.
        * function(location: Tensor) returning one of the above.

        Args:
          value: constant, Field, Tensor or function specifying the grid values
          type: type of Grid to create, must be either CenteredGrid or StaggeredGrid
          extrapolation: (optional) grid extrapolation, defaults to Domain.boundaries.scalar_extrapolation

        Returns:
          Grid of specified type
        """
        warnings.warn("Domain.grid is deprecated. Use scalar_grid or vector_grid instead.", DeprecationWarning)
        extrapolation = extrapolation or self.boundaries.scalar_extrapolation
        if type is CenteredGrid:
            return CenteredGrid.sample(value, self.resolution, self.bounds, extrapolation)
        elif type is StaggeredGrid:
            return StaggeredGrid.sample(value, self.resolution, self.bounds, extrapolation)
        else:
            raise ValueError('Unknown grid type: %s' % type)

    def scalar_grid(self,
             value: Field or Tensor or Number or Geometry or callable = 0,
             extrapolation: math.Extrapolation = None) -> CenteredGrid:
        """
        Creates a scalar grid matching the resolution and bounds of the domain.
        The grid is created from the given `value` which must be one of the following:

        * Number (int, float, complex or zero-dimensional tensor): all grid values will be equal to `value`. This has a near-zero memory footprint.
        * Scalar `Field`: the given value is resampled to the grid cells of this Domain.
        * Tensor with spatial dimensions matching the domain resolution: grid values will equal `value`.
        * Geometry: grid values are determined from the volume overlap between grid cells and geometry. Non-overlapping = 0, fully enclosed grid cell = 1.
        * function(location: Tensor) returning one of the above.
        * Native tensor: the number and order of axes are matched with the resolution of the domain.

        Args:
          value: constant, Field, Tensor or function specifying the grid values
          extrapolation: (optional) grid extrapolation, defaults to Domain.boundaries.scalar_extrapolation

        Returns:
          `CenteredGrid` with no channel dimensions
        """
        extrapolation = extrapolation or self.boundaries.scalar_extrapolation
        if isinstance(value, Field):
            assert_same_rank(value.spatial_rank, self.rank, f"Cannot resample {value.spatial_rank}D field to {self.rank}D domain.")
        elif isinstance(value, Tensor):
            assert value.shape.channel.rank == 0
        elif isinstance(value, (Number, Geometry)):
            pass
        else:
            value = math.tensor(value, names=self.resolution.names)
        result = CenteredGrid.sample(value, self.resolution, self.bounds, extrapolation)
        assert result.shape.channel_rank == 0
        return result

    def vector_grid(self,
                    value: Field or Tensor or Number or Geometry or callable = 0,
                    type: type = CenteredGrid,
                    extrapolation: math.Extrapolation = None) -> CenteredGrid or StaggeredGrid:
        """
        Creates a vector grid matching the resolution and bounds of the domain.
        The grid is created from the given `value` which must be one of the following:
        
        * Number (int, float, complex or zero-dimensional tensor): all grid values will be equal to `value`. This has a near-zero memory footprint.
        * Field: the given value is resampled to the grid cells of this Domain.
        * Tensor with spatial dimensions matcing the domain resolution: grid values will equal `value`.
        * Geometry: grid values are determined from the volume overlap between grid cells and geometry. Non-overlapping = 0, fully enclosed grid cell = 1.
        * function(location: Tensor) returning one of the above.
        
        The returned grid will have a vector dimension with size equal to the rank of the domain.

        Args:
          value: constant, Field, Tensor or function specifying the grid values
          type: class of Grid to create, must be either CenteredGrid or StaggeredGrid
          extrapolation: (optional) grid extrapolation, defaults to Domain.boundaries.scalar_extrapolation

        Returns:
          Grid of specified type

        """
        extrapolation = extrapolation or self.boundaries.vector_extrapolation
        if type is CenteredGrid:
            grid = CenteredGrid.sample(value, self.resolution, self.bounds, extrapolation)
            if grid.shape.channel.rank == 0:
                grid = grid.with_(values=math.expand_channel(grid.values, 'vector', dim_size=self.rank))
            else:
                assert grid.shape.channel.sizes[0] == self.rank
            return grid
        elif type is StaggeredGrid:
            return StaggeredGrid.sample(value, self.resolution, self.bounds, extrapolation)
        else:
            raise ValueError('Unknown grid type: %s' % type)

    def staggered_grid(self,
                       value: Field or Tensor or Number or Geometry or callable = 0,
                       extrapolation: math.Extrapolation = None) -> StaggeredGrid:
        """
        Creates a staggered grid matching the resolution and bounds of the domain.
        This is equal to calling `vector_grid()` with `type=StaggeredGrid`.
        
        The grid is created from the given `value` which must be one of the following:
        
        * Number (int, float, complex or zero-dimensional tensor): all grid values will be equal to `value`. This has a near-zero memory footprint.
        * Field: the given value is resampled to the grid cells of this Domain.
        * Tensor with spatial dimensions matcing the domain resolution: grid values will equal `value`.
        * Geometry: grid values are determined from the volume overlap between grid cells and geometry. Non-overlapping = 0, fully enclosed grid cell = 1.
        * function(location: Tensor) returning one of the above.
        
        The returned grid will have a vector dimension with size equal to the rank of the domain.

        Args:
          value: constant, Field, Tensor or function specifying the grid values
          type: class of Grid to create, must be either CenteredGrid or StaggeredGrid
          extrapolation: (optional) grid extrapolation, defaults to Domain.boundaries.scalar_extrapolation

        Returns:
          Grid of specified type
        """
        return self.vector_grid(value, type=StaggeredGrid, extrapolation=extrapolation)

    def accessible_mask(self, not_accessible: tuple or list, type: type = CenteredGrid) -> CenteredGrid or StaggeredGrid:
        """
        Unifies domain and Obstacle or Geometry objects into a binary StaggeredGrid mask which can be used
        to enforce boundary conditions.

        Args:
            not_accessible: blocked region(s) of space specified by geometries

        Returns:
            Binary mask indicating valid fields w.r.t. the boundary conditions.
        """
        accessible = HardGeometryMask(~union(not_accessible)) >> self.grid()
        accessible_mask = self.grid(accessible, extrapolation=self.boundaries.accessible_extrapolation)
        if type is CenteredGrid:
            return accessible_mask
        elif type is StaggeredGrid:
            return field.stagger(accessible_mask, math.minimum, self.boundaries.accessible_extrapolation)
        else:
            raise ValueError('Unknown grid type: %s' % type)

    def points(self,
               points: Tensor or Number or tuple or list,
               radius: Tensor or float or int or None = None,
               extrapolation: math.Extrapolation = None,
               color: str or Tensor or tuple or list or None = None) -> PointCloud:
        """
        Create a `phi.field.PointCloud` from the given `points`.
        The created field has no channel dimensions and all points carry the value `1`.

        Args:
            points: point locations in physical units
            radius: (optional) size of the particles
            extrapolation: (optional) extrapolation to use, defaults to extrapolation.ZERO
            color: (optional) color used when plotting the points

        Returns:
            `phi.field.PointCloud` object
        """
        extrapolation = extrapolation or math.extrapolation.ZERO
        if radius is None:
            radius = math.mean(self.bounds.size) * 0.005
        # --- Parse points: tuple / list ---
        if isinstance(points, (tuple, list)):
            if len(points) == 0:  # no points
                points = math.zeros(points=0, vector=1)
            elif isinstance(points[0], Number):  # single point
                points = math.tensor([points], 'points, vector')
            else:
                points = math.tensor(points, 'points, vector')
        elements = Sphere(points, radius)
        return PointCloud(elements, 1, extrapolation, add_overlapping=False, bounds=self.bounds, color=color)

    def distribute_points(self,
                          geometries: tuple or list,
                          points_per_cell: int = 8,
                          color: str = None,
                          center: bool = False) -> PointCloud:
        """
        Transforms `Geometry` objects into a PointCloud.

        Args:
            geometries: Geometry objects marking the cells which should contain points
            points_per_cell: Number of points for each cell of `geometries`
            color (Optional): Color of PointCloud
            center: Set all points to the center of the grid cells.

        Returns:
             PointCloud representation of `geometries`.
        """
        geometries = HardGeometryMask(union(geometries)) >> self.grid()
        initial_points = _distribute_points(geometries.values, points_per_cell, center=center)
        return self.points(initial_points, color=color)


@struct.definition()
class Obstacle(State):
    """
    An obstacle defines boundary conditions inside a geometry.
    It can also have a linear and angular velocity.
    """

    def __init__(self, geometry, material=CLOSED, velocity=0, tags=('obstacle',), **kwargs):
        State.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def geometry(self, geometry):
        """ Physical shape and size of the obstacle. """
        assert isinstance(geometry, Geometry)
        return geometry

    @struct.constant(default=CLOSED)
    def material(self, material):
        """ Boundary conditions to apply inside and on the surface of the obstacle. """
        assert isinstance(material, Material)
        return material

    @struct.constant(default=0)
    def velocity(self, velocity):
        """ Linear velocity vector of the obstacle. """
        return velocity

    @struct.constant(default=0)
    def angular_velocity(self, av):
        """ Rotation speed of the obstacle. Scalar value in 2D, vector in 3D. """
        return av

    @struct.derived()
    def is_stationary(self):
        """ Test whether the obstacle is completely still. """
        return self.velocity is 0 and self.angular_velocity is 0


class GeometryMovement(Physics):

    def __init__(self, geometry_function):
        Physics.__init__(self)
        self.geometry_at = geometry_function

    def step(self, obj, dt=1.0, **dependent_states):
        next_geometry = self.geometry_at(obj.age + dt)
        h = 1e-2 * dt if dt > 0 else 1e-2
        perturbed_geometry = self.geometry_at(obj.age + dt + h)
        velocity = (perturbed_geometry.center - next_geometry.center) / h
        if isinstance(obj, Obstacle):
            return obj.copied_with(geometry=next_geometry, velocity=velocity, age=obj.age + dt)
        if isinstance(obj, FieldEffect):
            with struct.ALL_ITEMS:
                next_field = struct.map(lambda x: x.copied_with(geometries=next_geometry) if isinstance(x, GeometryMask) else x, obj.field, leaf_condition=lambda x: isinstance(x, GeometryMask))
            return obj.copied_with(field=next_field, age=obj.age + dt)


def _distribute_points(mask: math.Tensor, points_per_cell: int = 1, center: bool = False) -> math.Tensor:
    """
    Generates points (either uniformly distributed or at the cell centers) according to the given tensor mask.

    Args:
        mask: Tensor with nonzero values at the indices where particles should get generated.
        points_per_cell: Number of particles to generate at each marked index
        center: Set points to cell centers. If False, points will be distributed using a uniform
            distribution within each cell.

    Returns:
        A tensor containing the positions of the generated points.
    """
    indices = math.to_float(math.nonzero(mask, list_dim='points'))
    temp = []
    for _ in range(points_per_cell):
        if center:
            temp.append(indices + 0.5)
        else:
            temp.append(indices + (math.random_uniform(indices.shape)))
    return math.concat(temp, dim='points')



import numpy as np

from phi import math
from phi.geom import GridCell, Geometry
from phi.math import random_normal, Tensor
from ._analytic import AnalyticField


class Noise(AnalyticField):

    def __init__(self, shape=math.EMPTY_SHAPE, scale=10, smoothness=1.0, **dims):
        """
        Generates random noise fluctuations which can be configured in physical size and smoothness.
            Each time values are sampled from a Noise field, a new noise field is generated.

            Noise is typically used as an initializer for CenteredGrids or StaggeredGrids.

        Args:
          channels: Number of independent random scalar fields this Field consists of
          scale: Size of noise fluctuations in physical units
          smoothness: Determines how quickly high frequencies die out
        """
        self.scale = scale
        self.smoothness = smoothness
        self._shape = shape & math.shape(**dims)

    @property
    def shape(self):
        return self._shape

    def sample_in(self, geometry: Geometry, reduce_channels=()) -> Tensor:
        if reduce_channels:
            shape = self._shape.non_channel.without(reduce_channels)
            assert len(reduce_channels) == 1
            geoms = geometry.unstack(reduce_channels[0])
            assert all(isinstance(g, GridCell) for g in geoms)
            components = [self.grid_sample(g.resolution, g.grid_size, shape) for g in geoms]
            return math.channel_stack(components, 'vector')
        if isinstance(geometry, GridCell):
            return self.grid_sample(geometry.resolution, geometry.grid_size)
        raise NotImplementedError(f"{type(geometry)} not supported. Only GridCell allowed.")

    def sample_at(self, points, reduce_channels=()) -> math.Tensor:
        raise NotImplementedError()

    def grid_sample(self, resolution: math.Shape, size, shape: math.Shape = None):
        shape = (self._shape if shape is None else shape).combined(resolution)
        rndj = math.to_complex(random_normal(shape)) + 1j * math.to_complex(random_normal(shape))  # Note: there is no complex32
        with math.SCIPY_BACKEND:
            k = math.fftfreq(resolution) * resolution / size * self.scale  # in physical units
            k = math.vec_squared(k)
        lowest_frequency = 0.1
        weight_mask = 1 / (1 + math.exp((lowest_frequency - k) * 1e3))  # High pass filter
        # --- Compute 1/k ---
        k.native()[(0,) * len(k.shape)] = np.inf
        inv_k = 1 / k
        inv_k.native()[(0,) * len(k.shape)] = 0
        # --- Compute result ---
        fft = rndj * inv_k ** self.smoothness * weight_mask
        array = math.real(math.ifft(fft))
        array /= math.std(array, dim=array.shape.non_batch)
        array -= math.mean(array, dim=array.shape.non_batch)
        array = math.to_float(array)
        return array

    def unstack(self, dimension: str) -> tuple:
        count = self.shape.get_size(dimension)
        reduced_shape = self.shape.without(dimension)
        return (Noise(reduced_shape, self.scale, self.smoothness),) * count

    def __repr__(self):
        return "%s, scale=%f, smoothness=%f" % (self._shape, self.scale, self.smoothness)



from phi import math
from ._geom import Geometry


def concat(geometries: tuple or list,
                   dim: str,
                   sizes: tuple or list or None = None):
    """
    Concatenates multiple geometries of the same type.

    Args:
        geometries: sequence of `phi.geom.Geometry` objects of the same type
        sizes: implicit
        dim: dimension to concatenate

    Returns:
        New `phi.geom.Geometry` object
    """
    if all(isinstance(g, type(geometries[0])) for g in geometries):
        characteristics = [g.__characteristics__() for g in geometries]
        if sizes is not None:
            characteristics = [{key: math.expand(val, dim, size) for key, val in c.items()}
                               for c, size in zip(characteristics, sizes)]
        new_attributes = {}
        for key in characteristics[0].keys():
            concatenated = math.concat([c[key] for c in characteristics], dim)
            new_attributes[key] = concatenated
        return geometries[0].__with__(**new_attributes)
    else:
        raise NotImplementedError()


def invert(geometry: Geometry):
    """
    Swaps inside and outside.

    Args:
        geometry: `phi.geom.Geometry` to swap

    Returns:
        New `phi.geom.Geometry` object with same surface but swapped normals
    """
    return ~geometry



# coding=utf-8
from __future__ import print_function

import inspect
import logging
import numbers
import os
import sys
import threading
import time
import warnings
from os.path import isfile

import numpy as np
from phi import struct, math
from phi.field import CenteredGrid, Field, StaggeredGrid, Scene
from phi.physics._world import StateProxy, world

from ._control import Action, Control
from ._value import (EditableBool, EditableFloat, EditableInt, EditableString, EditableValue)


def synchronized_method(method):
    outer_lock = threading.Lock()
    lock_name = '__' + method.__name__ + '_lock' + '__'

    def sync_method(self, *args, **kws):
        with outer_lock:
            if not hasattr(self, lock_name):
                setattr(self, lock_name, threading.Lock())
            lock = getattr(self, lock_name)
            with lock:
                return method(self, *args, **kws)

    return sync_method


class TimeDependentField(object):

    def __init__(self, name, generator):
        self.name = name
        self.generator = generator
        self.array = None
        self.invalidation_version = -1

    @synchronized_method
    def get(self, invalidation_version):
        if invalidation_version != self.invalidation_version:
            self.array = self.generator()
            self.invalidation_version = invalidation_version
        return self.array


class App(object):
    """
    Main class for defining an application that can be displayed in the user interface.
    
    To display data, call App.add_field().
    All fields need to be registered before the app is prepared or shown.
    
    To launch the GUI, call show(app). This calls App.prepare() if the app was not prepared.
    
    See the user interface documentation at https://tum-pbs.github.io/PhiFlow/Web_Interface.html

    Args:

    Returns:

    """

    def __init__(self,
                 name=None,
                 subtitle='',
                 fields=None,
                 stride=None,
                 base_dir='~/phi/data/',
                 summary=None,
                 custom_properties=None,
                 target_scene=None,
                 objects_to_save=None,
                 framerate=None,
                 dt=1.0):
        self.start_time = time.time()
        """ Time of creation (`App` constructor invocation) """
        self.name = name if name is not None else self.__class__.__name__
        """ Human-readable name. """
        self.subtitle = subtitle
        """ Description to be displayed. """
        self.summary = summary if summary else name
        """ The scene directory is derived from the summary. Defaults to `name`. """
        if fields:
            self.fields = {name: TimeDependentField(name, generator) for (name, generator) in fields.items()}
        else:
            self.fields = {}
        self.message = None
        self.steps = 0
        """ Counts the number of times `step()` has been called. May be set by the user. """
        self.time = 0
        """ Time variable for simulations. Can be set by the user. """
        self._invalidation_counter = 0
        self._controls = []
        self._actions = []
        self._traits = []
        self.prepared = False
        """ Wheter `prepare()` has been called. """
        self.current_action = None
        self._pause = False
        self.detect_fields = 'default'  # False, True, 'default'
        self.world = world
        self._dt = dt.initial_value if isinstance(dt, EditableValue) else dt
        if isinstance(dt, EditableValue):
            self._controls.append(Control(self, "dt", dt))
        self.min_dt = self._dt
        self.dt_history = {}  # sparse representation of time when new timestep was set (for the next step)
        # Setup directory & Logging
        self.objects_to_save = [self.__class__] if objects_to_save is None else list(objects_to_save)
        self.base_dir = os.path.expanduser(base_dir)
        if not target_scene:
            self.new_scene()
            self.uses_existing_scene = False
        else:
            self.scene = target_scene
            self.uses_existing_scene = True
        if not isfile(self.scene.subpath('info.log')):
            log_file = self.log_file = self.scene.subpath('info.log')
        else:
            index = 2
            while True:
                log_file = self.scene.subpath('info_%d.log' % index)
                if not isfile(log_file):
                    break
                else:
                    index += 1
        # Setup logging
        logFormatter = logging.Formatter('%(message)s (%(levelname)s), %(asctime)sn\n')
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logging.WARNING)
        customLogger = logging.Logger('app', logging.DEBUG)
        fileHandler = logging.FileHandler(log_file)
        fileHandler.setFormatter(logFormatter)
        customLogger.addHandler(fileHandler)
        consoleHandler = logging.StreamHandler(sys.stdout)
        consoleHandler.setFormatter(logFormatter)
        consoleHandler.setLevel(logging.INFO)
        customLogger.addHandler(consoleHandler)
        self.logger = customLogger
        # Framerate
        self.sequence_stride = stride if stride is not None else 1
        self.framerate = framerate if framerate is not None else stride
        """ Target frame rate in Hz. Play will not step faster than the framerate. `None` for unlimited frame rate. """
        self._custom_properties = custom_properties if custom_properties else {}
        # State
        self.state = None
        self.step_function = None
        # Initial log message
        self.info('App created. Scene directory is %s' % self.scene.path)

    @property
    def dt(self):
        """
        Current time increment per step.
        Used for `step_function` set by `set_state()` or for `world.step()` in legacy-style simulations. """
        return self._dt

    @dt.setter
    def dt(self, value):
        self._dt = value
        self.min_dt = min(self.min_dt, self.dt)
        self.dt_history[self.time] = self.dt

    def set_state(self, initial_state, step_function=None, show=(), dt=None):
        """
        Specifies the current physics state of the app and optionally the solver step function.
        The current physics state of the app is stored in `app.state`.
        
        This method replaces `world.add()` calls from Φ-Flow 1.

        Args:
          initial_state: dict mapping names (str) to Fields or Tensors
          step_function: function to progress the state. Called as step_function(dt=dt, **current_state) (Default value = None)
          show: list of names to expose to the user interface (Default value = ())
          dt: optional) value of dt to be passed to step_function (Default value = None)

        Returns:

        """
        self.state = initial_state
        self.step_function = step_function
        if dt is not None:
            self.dt = dt
        if show:
            if not self.prepared:
                for field_name in show:
                    self.add_field(field_name, lambda n=field_name: self.state[n])
                else:
                    warnings.warn('Ignoring show argument because App is already prepared.')

    @property
    def frame(self):
        """ Alias for `steps`. """
        return self.steps

    def new_scene(self, count=None):
        if count is None:
            count = 1 if self.world.batch_size is None else self.world.batch_size
        if count > 1:
            self.scene = Scene.create(os.path.join(self.base_dir, self.scene_summary()), batch=count)
        else:
            self.scene = Scene.create(os.path.join(self.base_dir, self.scene_summary()))

    @property
    def directory(self):
        """ This directory is automatically created upon `App` creation. Equal to `scene.path`. """
        return self.scene.path

    def _progress(self):
        # actual method called to step.
        self.step()
        self.steps += 1
        self.invalidate()

    def invalidate(self):
        """ Causes the user interface to update. """
        self._invalidation_counter += 1

    def step(self):
        """
        Performs a single step.
        You may override this method to specify what happens when the user presses the buttons `Step` or `Play`.
        
        If a step function has been passed to `App.set_state()`, the state is progressed using that function.
        
        Otherwise, `world.step()` is executed (for phiflow 1 style simulations).
        
        App.steps automatically counts how many steps have been completed.
        If this method is not overridden, `App.time` is additionally increased by `App.dt`.

        Args:

        Returns:

        """
        dt = self.dt  # prevent race conditions
        if self.step_function is None:
            world.step(dt=dt)
        else:
            new_state = self.step_function(dt=dt, **self.state)
            assert isinstance(self.state, dict), 'step_function must return a dict'
            assert new_state.keys() == self.state.keys(), 'step_function must return a state with the same names as the input state.\nInput: %s\nOutput: %s' % (self.state.keys(), new_state.keys())
            self.state = new_state
        self.time += dt

    @property
    def fieldnames(self):
        """ Alphabetical list of field names. See `get_field()`. """
        return sorted(self.fields.keys())

    def get_field(self, fieldname):
        """
        Reads the current value of a field.
        Fields can be added using `add_field()`.

        If a generator function was registered as the field data, this method may invoke the function which may take some time to complete.
        """
        if fieldname not in self.fields:
            raise KeyError('Field %s not declared. Available fields are %s' % (fieldname, self.fields.keys()))
        return self.fields[fieldname].get(self._invalidation_counter)

    def add_field(self, name: str, value):
        """
        Expose data to be displayed in the user interface.
        This method must be called before the user interface is launched, i.e. before `show(app)` or `app.prepare()` are invoked.
        
        `value` must be one of the following
        
        * Field
        * tensor
        * function without arguments returning one of the former. This function will be called each time the interface is updated.

        Args:
          name: unique human-readable name
          value: data to display
          name: str: 

        Returns:

        """
        assert not self.prepared, 'Cannot add fields to a prepared model'
        if isinstance(value, StateProxy):
            def current_state():
                return value.state
            generator = current_state
        elif callable(value):
            generator = value
        else:
            assert isinstance(value, (np.ndarray, Field, float, int, math.Tensor)), 'Unsupported type for field "%s": %s' % (name, type(value))

            def get_constant():
                return value
            generator = get_constant
        self.fields[name] = TimeDependentField(name, generator)

    @property
    def actions(self):
        """
        List of all custom actions that can be invoked at runtime by the user.
        Actions can be registered using `add_action()` or by defining a method with prefix `action_`.
        """
        return self._actions

    def add_action(self, name, methodcall):
        self._actions.append(Action(name, methodcall, name))

    def run_action(self, action):
        message_before = self.message
        action.method()
        self.invalidate()
        message_after = self.message
        if message_before == message_after:
            if self.message is None or self.message == '':
                self.message = display_name(action.name)
            else:
                self.message += ' | ' + display_name(action.name)

    @property
    def traits(self):
        return self._traits

    def add_trait(self, trait):
        assert not self.prepared, 'Cannot add traits to a prepared model'
        self._traits.append(trait)

    @property
    def controls(self):
        return self._controls

    def prepare(self):
        """
        Prepares the app to be displayed in a user interface.
        
        This method can only be called once.
        If not invoked manually, it is automatically called before the user interface is launched.
        
        Preparation includes:
        
        * Detecting editable values from member variables that start with 'value_'
        * Detecting actions from member functions that start with 'action_'
        * Initializing the scene directory with a JSON file and copying related Python source files
        
        :return: self

        Args:

        Returns:

        """
        if self.prepared:
            return
        logging.info('Gathering model data...')
        # Controls
        for name in self.__dict__:
            val = getattr(self, name)
            editable_value = None
            if isinstance(val, EditableValue):
                editable_value = val
                setattr(self, name, val.initial_value)  # Replace EditableValue with initial value
            elif name.startswith('value_'):
                value_name = display_name(name[6:])
                dtype = type(val)
                if dtype == bool:
                    editable_value = EditableBool(value_name, val)
                elif isinstance(val, numbers.Integral):  # Int
                    editable_value = EditableInt(value_name, val)
                elif isinstance(val, numbers.Number):  # Float
                    editable_value = EditableFloat(value_name, val)
                elif isinstance(val, str):
                    editable_value = EditableString(value_name, val)
            if editable_value:
                self._controls.append(Control(self, name, editable_value))
        # Actions
        for method_name in dir(self):
            if method_name.startswith('action_') and callable(getattr(self, method_name)):
                self._actions.append(Action(display_name(method_name[7:]), getattr(self, method_name), method_name))
        # Default fields
        if len(self.fields) == 0:
            self._add_default_fields()
        # Scene
        self._update_scene_properties()
        source_files_to_save = set()
        for object in self.objects_to_save:
            try:
                source_files_to_save.add(inspect.getabsfile(object))
            except TypeError:
                pass
        for source_file in source_files_to_save:
            self.scene.copy_src(source_file)
        # End
        self.prepared = True
        return self

    def _add_default_fields(self):
        def add_default_field(trace):
            field = trace.value
            if isinstance(field, (CenteredGrid, StaggeredGrid)):
                def field_generator():
                    world_state = self.world.state
                    return trace.find_in(world_state)
                self.add_field(field.name[0].upper() + field.name[1:], field_generator)
            return None
        struct.map(add_default_field, self.world.state, leaf_condition=lambda x: isinstance(x, (CenteredGrid, StaggeredGrid)), trace=True, content_type=struct.INVALID)

    def add_custom_property(self, key, value):
        self._custom_properties[key] = value
        if self.prepared:
            self._update_scene_properties()

    def add_custom_properties(self, dictionary):
        self._custom_properties.update(dictionary)
        if self.prepared:
            self._update_scene_properties()

    def _update_scene_properties(self):
        if self.uses_existing_scene:
            return
        try:
            app_name = os.path.basename(inspect.getfile(self.__class__))
            app_path = inspect.getabsfile(self.__class__)
        except TypeError:
            app_name = app_path = ''
        properties = {
            'instigator': 'App',
            'traits': self.traits,
            'app': str(app_name),
            'app_path': str(app_path),
            'name': self.name,
            'description': self.subtitle,
            'all_fields': self.fieldnames,
            'actions': [action.name for action in self.actions],
            'controls': [{control.name: control.value} for control in self.controls],
            'summary': self.scene_summary(),
            'steps': self.steps,
            'time': self.time,
            'world': struct.properties_dict(self.world.state)
        }
        properties.update(self.custom_properties())
        self.scene.properties = properties

    def settings_str(self):
        return ''.join([
            ' ' + str(control) for control in self.controls
        ])

    def custom_properties(self):
        return self._custom_properties

    def info(self, message: str):
        """
        Update the status message.
        The status message is written to the console and the log file.
        Additionally, it may be displayed by the user interface.

        See `debug()`.

        Args:
            message: Message to display
        """
        message = str(message)
        self.message = message
        self.logger.info(message)

    def debug(self, message):
        """
        Prints a message to the log file but does not display it.

        See `info()`.

        Args:
            message: Message to log.
        """
        logging.info(message)

    def scene_summary(self):
        return self.summary

    def show(self, **config):
        warnings.warn("Use show(model) instead.", DeprecationWarning, stacklevel=2)
        from ._display import show
        show(self, **config)

    @property
    def status(self):
        pausing = '/Pausing' if (self._pause and self.current_action) else ''
        action = self.current_action if self.current_action else 'Idle'
        message = f' - {self.message}' if self.message else ''
        return f'{action}{pausing} (t={self.format_time(self.time)} in {self.steps} steps){message}'

    def format_time(self, time):
        commas = int(np.ceil(np.abs(np.log10(self.min_dt))))
        return ("{time:," + f".{commas}f" + "}").format(time=time)

    def run_step(self, framerate=None):
        self.current_action = 'Running'
        starttime = time.time()
        try:
            self._progress()
            if framerate is not None:
                duration = time.time() - starttime
                rest = 1.0 / framerate - duration
                if rest > 0:
                    self.current_action = 'Waiting'
                    time.sleep(rest)
        except Exception as e:
            self.info('Error during %s.step() \n %s: %s' % (type(self).__name__, type(e).__name__, e))
            self.logger.exception(e)
        finally:
            self.current_action = None

    def play(self, max_steps=None, callback=None, framerate=None, callback_if_aborted=False):
        """
        Run a number of steps.

        Args:
            max_steps: (optional) stop when this many steps have been completed (independent of the `steps` variable) or `pause()` is called.
            callback: Function to be run after all steps have been completed.
            framerate: Target frame rate in Hz.
            callback_if_aborted: Whether to invoke `callback` if `pause()` causes this method to abort prematurely.

        Returns:
            self
        """
        if framerate is None:
            framerate = self.framerate

        def target():
            self._pause = False
            step_count = 0
            while not self._pause:
                self.run_step(framerate=framerate)
                step_count += 1
                if max_steps and step_count >= max_steps:
                    break
            if callback is not None:
                if not self._pause or callback_if_aborted:
                    callback()

        thread = threading.Thread(target=target)
        thread.start()
        return self

    def pause(self):
        """ Causes the `play()` method to stop after finishing the current step. """
        self._pause = True

    @property
    def running(self):
        """ Whether `play()` is currently executing. """
        return self.current_action is not None

    def benchmark(self, sequence_count):
        self._pause = False
        step_count = 0
        starttime = time.time()
        for i in range(sequence_count):
            self.run_step(framerate=np.inf)
            step_count += 1
            if self._pause:
                break
        time_elapsed = time.time() - starttime
        return step_count, time_elapsed


def display_name(python_name):
    n = list(python_name)
    n[0] = n[0].upper()
    for i in range(1, len(n)):
        if n[i] == '_':
            n[i] = ' '
            if len(n) > i + 1:
                n[i + 1] = n[i + 1].upper()
    return ''.join(n)



from typing import Any

from phi import math
from phi.geom import Geometry, GridCell, Box
from ._field import SampledField
from ..geom._stack import GeometryStack
from ..math import Tensor


class PointCloud(SampledField):

    def __init__(self, elements: Geometry,
                 values: Any = 1,
                 extrapolation=math.extrapolation.ZERO,
                 add_overlapping=False,
                 bounds: Box = None,
                 color: str or Tensor or tuple or list or None = None):
        """
        A point cloud consists of elements at arbitrary locations.
        A value or vector is associated with each element.

        Outside of elements, the value of the field is determined by the extrapolation.

        All points belonging to one example must be listed in the 'points' dimension.

        Unlike with GeometryMask, the elements of a PointCloud are assumed to be small.
        When sampling this field on a grid, scatter functions may be used.

        See the `phi.field` module documentation at https://tum-pbs.github.io/PhiFlow/Fields.html

        Args:
          elements: Geometry object specifying the sample points and sizes
          values: values corresponding to elements
          extrapolation: values outside elements
          add_overlapping: True: values of overlapping geometries are summed. False: values between overlapping geometries are interpolated
          bounds: (optional) size of the fixed domain in which the points should get visualized. None results in max and min coordinates of points.
          color: (optional) hex code for color or tensor of colors (same length as elements) in which points should get plotted.
        """
        SampledField.__init__(self, elements, values, extrapolation)
        self._add_overlapping = add_overlapping
        assert bounds is None or isinstance(bounds, Box), 'Invalid bounds.'
        self._bounds = bounds
        assert 'points' in self.shape, "Cannot create PointCloud without 'points' dimension. Add it either to elements or to values as batch dimension."
        if color is None:
            color = '#0060ff'
        self._color = math.tensor(color, names='points') if isinstance(color, (tuple, list)) else math.tensor(color)

    @property
    def bounds(self) -> Box:
        return self._bounds

    @property
    def color(self) -> Tensor:
        return self._color

    def sample_in(self, geometry: Geometry, reduce_channels=()) -> Tensor:
        if not reduce_channels:
            if geometry == self.elements:
                return self.values
            elif isinstance(geometry, GridCell):
                return self._grid_scatter(geometry.bounds, geometry.resolution)
            elif isinstance(geometry, GeometryStack):
                sampled = [self.sample_at(g) for g in geometry.geometries]
                return math.batch_stack(sampled, geometry.stack_dim_name)
            else:
                raise NotImplementedError()
        else:
            assert len(reduce_channels) == 1
            components = self.unstack('vector') if 'vector' in self.shape else (self,) * geometry.shape.get_size(reduce_channels[0])
            sampled = [c.sample_in(p) for c, p in zip(components, geometry.unstack(reduce_channels[0]))]
            return math.channel_stack(sampled, 'vector')

    def sample_at(self, points, reduce_channels=()) -> Tensor:
        raise NotImplementedError()

    def _grid_scatter(self, box: Box, resolution: math.Shape):
        """
        Approximately samples this field on a regular grid using math.scatter().

        Args:
          box: physical dimensions of the grid
          resolution: grid resolution
          box: Box: 
          resolution: math.Shape: 

        Returns:
          CenteredGrid

        """
        closest_index = math.to_int(math.round(box.global_to_local(self.points) * resolution - 0.5))
        if self._add_overlapping:
            duplicates_handling = 'add'
        else:
            duplicates_handling = 'mean'
        scattered = math.scatter(closest_index, self.values, resolution, duplicates_handling=duplicates_handling, outside_handling='discard', scatter_dims=('points',))
        return scattered

    def __repr__(self):
        return "PointCloud[%s]" % (self.shape,)

    def __and__(self, other):
        assert isinstance(other, PointCloud)
        from ._field_math import concat
        return concat(self, other, dim='points')



import numpy as np


class AxisOrder:
    """Define the order of spatial axes. Default: x first"""

    def __init__(self):
        self.is_x_first = False
        self.x = 0
        self.y = 0
        self.z = 0
        self.x_first()

    def x_first(self):
        self.is_x_first = True
        self.x, self.y, self.z = 0, 1, 2

    def x_last(self):
        self.is_x_first = False
        self.x, self.y, self.z = -1, -2, -3

    def up_vector(self, rank):
        if self.is_x_first:
            return np.array([0] * (rank - 1) + [1])
        else:
            return np.array([1] + [0] * (rank - 1))

    def axis_name(self, index, spatial_rank):
        if self.is_x_first:
            return ['x', 'y', 'z', 'w'][:spatial_rank][index]
        else:
            return ['x', 'y', 'z', 'w'][:spatial_rank][::-1][index]


GLOBAL_AXIS_ORDER = AxisOrder()



import dash_core_components as dcc
import dash_html_components as html

from .board import build_benchmark, build_tf_profiler, build_tensorboard_launcher, build_system_controls
from .log import build_log
from .model_controls import build_model_controls
from .viewsettings import build_view_selection
from .dash_app import DashApp
from .info import build_app_details, build_description, build_phiflow_info, build_app_time
from .viewer import build_viewer
from .player_controls import build_status_bar, build_player_controls
from .._display import AppDisplay
from .._display_util import ordered_field_names


class DashGui(AppDisplay):

    def __init__(self, app):
        AppDisplay.__init__(self, app)
        self.dash_app = None

    def setup(self):
        header_layout = html.Div([
                dcc.Link('Home', href='/'),
                ' - ',
                dcc.Link('Side-by-Side', href='/side-by-side'),
                ' - ',
                dcc.Link('Quad', href='/quad'),
                ' - ',
                dcc.Link('Info', href='/info'),
                ' - ',
                dcc.Link('Log', href='/log'),
                ' - ',
                dcc.Link(u'Φ Board', href='/board'),
                # ' - ',
                # dcc.Link('Scripting', href='/scripting'),
                ' - ',
                html.A('Help', href='https://tum-pbs.github.io/PhiFlow/Web_Interface.html', target='_blank'),
            ])
        dash_app = self.dash_app = DashApp(self.app, self.config, header_layout)

        disp_fields = ordered_field_names(self.app, self.config.get('display'), min_count=4, fill_with='None')

        # --- Shared components ---
        player_controls = build_player_controls(dash_app)
        status_bar = build_status_bar(dash_app)
        model_controls = build_model_controls(dash_app)

        # --- Home ---
        layout = html.Div([
            build_description(dash_app),
            build_view_selection(dash_app),
            html.Div(style={'width': 1000, 'height': 800, 'margin-left': 'auto', 'margin-right': 'auto'}, children=[
                build_viewer(dash_app, 800, id='home', initial_field_name=disp_fields[0], config=self.config),
            ]),
            status_bar,
            player_controls,
            model_controls,
        ])
        dash_app.add_page('/', layout)

        # --- Side by Side ---
        layout = html.Div([
            build_view_selection(dash_app),
            html.Div(style={'width': '50%', 'height': 700, 'display': 'inline-block'}, children=[
                build_viewer(dash_app, 700, id='left', initial_field_name=disp_fields[0], config=self.config),
            ]),
            html.Div(style={'width': '50%', 'height': 700, 'display': 'inline-block'}, children=[
                build_viewer(dash_app, 700, id='right', initial_field_name=disp_fields[1], config=self.config),
            ]),
            status_bar,
            player_controls,
            model_controls,
        ])
        dash_app.add_page('/side-by-side', layout)

        # --- Quad ---
        layout = html.Div([
            build_view_selection(dash_app),
            html.Div(style={'width': '50%', 'height': 700, 'display': 'inline-block'}, children=[
                build_viewer(dash_app, 700, id='top-left', initial_field_name=disp_fields[0], config=self.config),
            ]),
            html.Div(style={'width': '50%', 'height': 700, 'display': 'inline-block'}, children=[
                build_viewer(dash_app, 700, id='top-right', initial_field_name=disp_fields[1], config=self.config),
            ]),
            html.Div(style={'width': '50%', 'height': 700, 'display': 'inline-block'}, children=[
                build_viewer(dash_app, 700, id='bottom-left', initial_field_name=disp_fields[2], config=self.config),
            ]),
            html.Div(style={'width': '50%', 'height': 700, 'display': 'inline-block'}, children=[
                build_viewer(dash_app, 700, id='bottom-right', initial_field_name=disp_fields[3], config=self.config),
            ]),
            status_bar,
            player_controls,
            model_controls,
        ])
        dash_app.add_page('/quad', layout)

        # --- Log ---
        layout = html.Div([
            dcc.Markdown('# Log'),
            status_bar,
            player_controls,
            build_log(dash_app)
        ])
        dash_app.add_page('/log', layout)

        # --- Info ---
        layout = html.Div([
            build_description(dash_app),
            status_bar,
            player_controls,
            build_phiflow_info(dash_app),
            build_app_details(dash_app),
            build_app_time(dash_app),
        ])
        dash_app.add_page('/info', layout)

        # --- Board ---
        layout = html.Div([
            dcc.Markdown(u'# Φ Board'),
            status_bar,
            player_controls,
        ] + ([] if 'tensorflow' not in dash_app.app.traits else [
            build_tensorboard_launcher(dash_app),
        ]) + [
            model_controls,
            build_benchmark(dash_app),
        ] + ([] if 'tensorflow' not in dash_app.app.traits else [
            build_tf_profiler(dash_app),
        ]) + [
            build_system_controls(dash_app),
            # ToDo: Graphs, Record/Animate
        ])
        dash_app.add_page('/board', layout)

        # --- Scripting ---
        layout = html.Div([
            dcc.Markdown(u'# Python Scripting'),
            'Custom Fields, Execute script, Restart'
        ])
        dash_app.add_page('/scripting', layout)

        return self.dash_app.dash

    def show(self, caller_is_main):
        if not caller_is_main and self.config.get('external_web_server', False):
            return self.dash_app.server
        else:
            port = self.config.get('port', 8051)
            print('Starting Dash server on http://localhost:%d/' % port)
            self.dash_app.dash.run_server(debug=True, host='0.0.0.0', port=port, use_reloader=False)
            return self



import numbers
import os
import sys
import warnings
from typing import List

import numpy as np
import scipy.signal
import scipy.sparse
from scipy.sparse.linalg import cg, LinearOperator

from . import Backend, ComputeDevice
from ._backend_helper import combined_dim
from ._dtype import from_numpy_dtype, to_numpy_dtype, DType
from ._optim import Solve, LinearSolve, SolveResult


class SciPyBackend(Backend):

    """Core Python Backend using NumPy & SciPy"""

    def __init__(self):
        Backend.__init__(self, "SciPy")

    def list_devices(self, device_type: str or None = None) -> List[ComputeDevice]:
        if sys.platform != "win32":
            mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')
        else:
            mem_bytes = -1
        processors = os.cpu_count()
        return [ComputeDevice("CPU", 'CPU', mem_bytes, processors, "")]

    def as_tensor(self, x, convert_external=True):
        if self.is_tensor(x, only_native=convert_external):
            array = x
        else:
            array = np.array(x)
        # --- Enforce Precision ---
        if not isinstance(array, numbers.Number):
            if array.dtype in (np.float16, np.float32, np.float64, np.longdouble):
                array = self.to_float(array)
        return array

    def is_tensor(self, x, only_native=False):
        if isinstance(x, np.ndarray) and x.dtype != np.object:
            return True
        if scipy.sparse.issparse(x):
            return True
        if isinstance(x, np.bool_):
            return True
        # --- Above considered native ---
        if only_native:
            return False
        # --- Non-native types
        if isinstance(x, (numbers.Number, bool, str)):
            return True
        if isinstance(x, (tuple, list)):
            return all([self.is_tensor(item, False) for item in x])
        return False

    def is_available(self, tensor):
        return True

    def numpy(self, tensor):
        if isinstance(tensor, np.ndarray):
            return tensor
        else:
            return np.array(tensor)

    def copy(self, tensor, only_mutable=False):
        return np.copy(tensor)

    def trace_function(self, f: callable) -> callable:
        return f

    def custom_gradient(self, f: callable, gradient: callable) -> callable:
        warnings.warn("SciPy backend does not support gradients, custom gradient will be ignored.")
        return f

    def transpose(self, tensor, axes):
        return np.transpose(tensor, axes)

    def equal(self, x, y):
        if isinstance(x, np.ndarray) and x.dtype.char == 'U':  # string comparison
            x = x.astype(np.object)
        if isinstance(x, str):
            x = np.array(x, np.object)
        return np.equal(x, y)

    def divide_no_nan(self, x, y):
        with np.errstate(divide='ignore', invalid='ignore'):
            result = x / y
        return np.where(y == 0, 0, result)

    def random_uniform(self, shape):
        return np.random.random(shape).astype(to_numpy_dtype(self.float_type))

    def random_normal(self, shape):
        return np.random.standard_normal(shape).astype(to_numpy_dtype(self.float_type))

    def range(self, start, limit=None, delta=1, dtype=None):
        """
        range syntax to arange syntax

        Args:
          start: 
          limit:  (Default value = None)
          delta:  (Default value = 1)
          dtype:  (Default value = None)

        Returns:

        """
        if limit is None:
            start, limit = 0, start
        return np.arange(start, limit, delta, dtype)

    def tile(self, value, multiples):
        return np.tile(value, multiples)

    def stack(self, values, axis=0):
        return np.stack(values, axis)

    def concat(self, values, axis):
        return np.concatenate(values, axis)

    def pad(self, value, pad_width, mode='constant', constant_values=0):
        assert mode in ('constant', 'symmetric', 'periodic', 'reflect', 'boundary'), mode
        if mode == 'constant':
            return np.pad(value, pad_width, 'constant', constant_values=constant_values)
        else:
            if mode in ('periodic', 'boundary'):
                mode = {'periodic': 'wrap', 'boundary': 'edge'}[mode]
            return np.pad(value, pad_width, mode)

    def reshape(self, value, shape):
        return np.reshape(value, shape)

    def sum(self, value, axis=None, keepdims=False):
        return np.sum(value, axis=axis, keepdims=keepdims)

    def prod(self, value, axis=None):
        if not isinstance(value, np.ndarray):
            value = np.array(value)
        if value.dtype == bool:
            return np.all(value, axis=axis)
        return np.prod(value, axis=axis)

    def where(self, condition, x=None, y=None):
        if x is None or y is None:
            return np.argwhere(condition)
        return np.where(condition, x, y)

    def nonzero(self, values):
        return np.argwhere(values)

    def py_func(self, func, inputs, Tout, shape_out, stateful=True, name=None, grad=None):
        result = func(*inputs)
        assert result.dtype == Tout, "returned value has wrong type: {}, expected {}".format(result.dtype, Tout)
        assert result.shape == shape_out, "returned value has wrong shape: {}, expected {}".format(result.shape, shape_out)
        return result

    def zeros(self, shape, dtype: DType = None):
        return np.zeros(shape, dtype=to_numpy_dtype(dtype or self.float_type))

    def zeros_like(self, tensor):
        return np.zeros_like(tensor)

    def ones(self, shape, dtype: DType = None):
        return np.ones(shape, dtype=to_numpy_dtype(dtype or self.float_type))

    def ones_like(self, tensor):
        return np.ones_like(tensor)

    def meshgrid(self, *coordinates):
        return np.meshgrid(*coordinates, indexing='ij')

    def linspace(self, start, stop, number):
        return np.linspace(start, stop, number, dtype=to_numpy_dtype(self.float_type))

    def mean(self, value, axis=None, keepdims=False):
        return np.mean(value, axis, keepdims=keepdims)

    def dot(self, a, b, axes):
        return np.tensordot(a, b, axes)

    def mul(self, a, b):
        if scipy.sparse.issparse(a):
            return a.multiply(b)
        elif scipy.sparse.issparse(b):
            return b.multiply(a)
        else:
            return Backend.mul(self, a, b)

    def matmul(self, A, b):
        return np.stack([A.dot(b[i]) for i in range(b.shape[0])])

    def einsum(self, equation, *tensors):
        return np.einsum(equation, *tensors)

    def while_loop(self, cond, body, loop_vars, shape_invariants=None, parallel_iterations=10, back_prop=True,
                   swap_memory=False, name=None, maximum_iterations=None):
        i = 0
        while cond(*loop_vars):
            if maximum_iterations is not None and i == maximum_iterations:
                break
            loop_vars = body(*loop_vars)
            i += 1
        return loop_vars

    def abs(self, x):
        return np.abs(x)

    def sign(self, x):
        return np.sign(x)

    def round(self, x):
        return np.round(x)

    def ceil(self, x):
        return np.ceil(x)

    def floor(self, x):
        return np.floor(x)

    def max(self, x, axis=None, keepdims=False):
        return np.max(x, axis, keepdims=keepdims)

    def min(self, x, axis=None, keepdims=False):
        return np.min(x, axis, keepdims=keepdims)

    def maximum(self, a, b):
        return np.maximum(a, b)

    def minimum(self, a, b):
        return np.minimum(a, b)

    def clip(self, x, minimum, maximum):
        return np.clip(x, minimum, maximum)

    def sqrt(self, x):
        return np.sqrt(x)

    def exp(self, x):
        return np.exp(x)

    def conv(self, tensor, kernel, padding="SAME"):
        """
        apply convolution of kernel on tensor

        Args:
          tensor: 
          kernel: 
          padding:  (Default value = "SAME")

        Returns:

        """
        assert tensor.shape[-1] == kernel.shape[-2]
        # kernel = kernel[[slice(None)] + [slice(None, None, -1)] + [slice(None)]*(len(kernel.shape)-3) + [slice(None)]]
        if padding.lower() == "same":
            result = np.zeros(tensor.shape[:-1] + (kernel.shape[-1],), dtype=to_numpy_dtype(self.float_type))
        elif padding.lower() == "valid":
            valid = [tensor.shape[i + 1] - (kernel.shape[i] + 1) // 2 for i in range(tensor_spatial_rank(tensor))]
            result = np.zeros([tensor.shape[0]] + valid + [kernel.shape[-1]], dtype=to_numpy_dtype(self.float_type))
        else:
            raise ValueError("Illegal padding: %s" % padding)
        for batch in range(tensor.shape[0]):
            for o in range(kernel.shape[-1]):
                for i in range(tensor.shape[-1]):
                    result[batch, ..., o] += scipy.signal.correlate(tensor[batch, ..., i], kernel[..., i, o], padding.lower())
        return result

    def expand_dims(self, a, axis=0, number=1):
        for _i in range(number):
            a = np.expand_dims(a, axis)
        return a

    def shape(self, tensor):
        return np.shape(tensor)

    def staticshape(self, tensor):
        return np.shape(tensor)

    def cast(self, x, dtype: DType):
        if self.is_tensor(x, only_native=True) and from_numpy_dtype(x.dtype) == dtype:
            return x
        else:
            return np.array(x, to_numpy_dtype(dtype))

    def gather(self, values, indices):
        if scipy.sparse.issparse(values):
            if scipy.sparse.isspmatrix_coo(values):
                values = values.tocsc()
        return values[indices]

    def gather_nd(self, values, indices, batch_dims=0):
        assert indices.shape[-1] == self.ndims(values) - batch_dims - 1
        if batch_dims == 0:
            indices_list = self.unstack(indices, axis=-1)
            result = values[indices_list]
            return result
        for dim in range(batch_dims):
            assert indices.shape[dim] == values.shape[dim] or values.shape[dim] == 1 or indices.shape[dim] == 1, 'Batch dimension %d does not match: %s (values) and %s (indices)' % (dim, values.shape, indices.shape)
        values_batch_max = np.array(values.shape[:batch_dims]) - 1
        indices_batch_max = np.array(indices.shape[:batch_dims]) - 1

        def inner_gather_nd(*pos):
            batch_idx_values = tuple([np.minimum(pos[i], values_batch_max[i]) for i in range(len(pos))])
            values_batch = values[batch_idx_values]
            batch_idx_indices = tuple([np.minimum(pos[i], indices_batch_max[i]) for i in range(len(pos))])
            indices_batch = indices[batch_idx_indices]
            result = values_batch[self.unstack(indices_batch, axis=-1)]
            return result
        # --- Iterate over batch dimensions ---
        batch_pos = np.meshgrid(*[range(dim) for dim in indices.shape[:batch_dims]], indexing='ij')
        batch_pos = np.stack(batch_pos, axis=-1).reshape([-1, batch_dims])
        result = np.empty(indices.shape[:-1] + (values.shape[-1],), values.dtype)
        for i in range(batch_pos.shape[0]):
            gathered = inner_gather_nd(*batch_pos[i])
            result[tuple(batch_pos[i])] = gathered
        return result

    def std(self, x, axis=None, keepdims=False):
        return np.std(x, axis, keepdims=keepdims)

    def boolean_mask(self, x, mask):
        return x[mask]

    def isfinite(self, x):
        return np.isfinite(x)

    def any(self, boolean_tensor, axis=None, keepdims=False):
        return np.any(boolean_tensor, axis=axis, keepdims=keepdims)

    def all(self, boolean_tensor, axis=None, keepdims=False):
        return np.all(boolean_tensor, axis=axis, keepdims=keepdims)

    def scatter(self, indices, values, shape, duplicates_handling='undefined', outside_handling='undefined'):
        assert duplicates_handling in ('undefined', 'add', 'mean', 'any')
        assert outside_handling in ('discard', 'clamp', 'undefined')
        shape = np.array(shape, np.int32)
        if outside_handling == 'clamp':
            indices = np.maximum(0, np.minimum(indices, shape - 1))
        elif outside_handling == 'discard':
            indices_inside = (indices >= 0) & (indices < shape)
            indices_inside = np.min(indices_inside, axis=-1)
            filter_indices = np.argwhere(indices_inside)
            indices = indices[filter_indices][..., 0, :]
            if values.shape[0] > 1:
                values = values[filter_indices.reshape(-1)]
        array = np.zeros(tuple(shape) + values.shape[indices.ndim-1:], to_numpy_dtype(self.float_type))
        indices = self.unstack(indices, axis=-1)
        if duplicates_handling == 'add':
            np.add.at(array, tuple(indices), values)
        elif duplicates_handling == 'mean':
            count = np.zeros(shape, np.int32)
            np.add.at(array, tuple(indices), values)
            np.add.at(count, tuple(indices), 1)
            count = np.maximum(1, count)
            return array / count
        else:  # last, any, undefined
            array[indices] = values
        return array

    def fft(self, x):
        rank = len(x.shape) - 2
        assert rank >= 1
        if rank == 1:
            return np.fft.fft(x, axis=1)
        elif rank == 2:
            return np.fft.fft2(x, axes=[1, 2])
        else:
            return np.fft.fftn(x, axes=list(range(1, rank + 1)))

    def ifft(self, k):
        assert self.dtype(k).kind == complex
        rank = len(k.shape) - 2
        assert rank >= 1
        if rank == 1:
            return np.fft.ifft(k, axis=1).astype(k.dtype)
        elif rank == 2:
            return np.fft.ifft2(k, axes=[1, 2]).astype(k.dtype)
        else:
            return np.fft.ifftn(k, axes=list(range(1, rank + 1))).astype(k.dtype)

    def imag(self, complex_arr):
        return np.imag(complex_arr)

    def real(self, complex_arr):
        return np.real(complex_arr)

    def sin(self, x):
        return np.sin(x)

    def cos(self, x):
        return np.cos(x)

    def dtype(self, array) -> DType:
        if isinstance(array, int):
            return DType(int, 32)
        if isinstance(array, float):
            return DType(float, 64)
        if isinstance(array, complex):
            return DType(complex, 128)
        if not isinstance(array, np.ndarray):
            array = np.array(array)
        return from_numpy_dtype(array.dtype)

    def sparse_tensor(self, indices, values, shape):
        if not isinstance(indices, (tuple, list)):
            indices = self.unstack(indices, -1)
        if len(indices) == 2:
            return scipy.sparse.csc_matrix((values, indices), shape=shape)
        else:
            raise NotImplementedError(f"len(indices) = {len(indices)} not supported. Only (2) allowed.")

    def coordinates(self, tensor, unstack_coordinates=False):
        if scipy.sparse.issparse(tensor):
            coo = tensor.tocoo()
            return (coo.row, coo.col), coo.data
        else:
            raise NotImplementedError("Only sparse tensors supported.")

    def conjugate_gradient(self, A, y, x0, solve_params=LinearSolve(), gradient: str = 'implicit', callback=None):
        bs_y = self.staticshape(y)[0]
        bs_x0 = self.staticshape(x0)[0]
        batch_size = combined_dim(bs_y, bs_x0)

        if callable(A):
            A = LinearOperator(dtype=y.dtype, shape=(self.staticshape(y)[-1], self.staticshape(x0)[-1]), matvec=A)
        elif isinstance(A, (tuple, list)) or self.ndims(A) == 3:
            batch_size = combined_dim(batch_size, self.staticshape(A)[0])

        iterations = [0] * batch_size
        converged = []
        results = []

        def count_callback(*args):
            iterations[batch] += 1
            if callback is not None:
                callback(*args)

        for batch in range(batch_size):
            y_ = y[min(batch, bs_y - 1)]
            x0_ = x0[min(batch, bs_x0 - 1)]
            x, ret_val = cg(A, y_, x0_, tol=solve_params.relative_tolerance, atol=solve_params.absolute_tolerance, maxiter=solve_params.max_iterations, callback=count_callback)
            converged.append(ret_val == 0)
            results.append(x)
        solve_params.result = SolveResult(all(converged), max(iterations))
        return self.stack(results)


def clamp(coordinates, shape):
    assert coordinates.shape[-1] == len(shape)
    for i in range(len(shape)):
        coordinates[...,i] = np.maximum(0, np.minimum(shape[i] - 1, coordinates[..., i]))
    return coordinates


def tensor_spatial_rank(field):
    dims = len(field.shape) - 2
    assert dims > 0, "channel has no spatial dimensions"
    return dims


SCIPY_BACKEND = SciPyBackend()



from functools import wraps
from typing import TypeVar, Tuple

from phi import math
from phi import geom
from phi.geom import Box, Geometry
from phi.math import extrapolate_valid_values
from ._field import Field, SampledField
from ._grid import CenteredGrid, Grid, StaggeredGrid
from ._point_cloud import PointCloud
from ._mask import HardGeometryMask


def laplace(field: Grid, axes=None):
    result = field._op1(lambda tensor: math.laplace(tensor, dx=field.dx, padding=field.extrapolation, dims=axes))
    return result


def gradient(field: CenteredGrid, type: type = CenteredGrid, stack_dim='vector'):
    """
    Finite difference gradient.

    This function can operate in two modes:

    * `type=CenteredGrid` approximates the gradient at cell centers using central differences
    * `type=StaggeredGrid` computes the gradient at face centers of neighbouring cells

    Args:
        field: centered grid of any number of dimensions (scalar field, vector field, tensor field)
        type: either `CenteredGrid` or `StaggeredGrid`
        stack_dim: name of dimension to be added. This dimension lists the gradient w.r.t. the spatial dimensions.
            The `field` must not have a dimension of the same name.

    Returns:
        gradient field of type `type`.

    """
    if type == CenteredGrid:
        values = math.gradient(field.values, field.dx.vector.as_channel(name=stack_dim), difference='central', padding=field.extrapolation, stack_dim=stack_dim)
        return CenteredGrid(values, field.bounds, field.extrapolation.gradient())
    elif type == StaggeredGrid:
        assert stack_dim == 'vector'
        return stagger(field, lambda lower, upper: (upper - lower) / field.dx, field.extrapolation.gradient())
    raise NotImplementedError(f"{type(field)} not supported. Only CenteredGrid and StaggeredGrid allowed.")


def shift(grid: CenteredGrid, offsets: tuple, stack_dim='shift'):
    """
    Wraps :func:`math.shift` for CenteredGrid.

    Args:
      grid: CenteredGrid: 
      offsets: tuple: 
      stack_dim:  (Default value = 'shift')

    Returns:

    """
    data = math.shift(grid.values, offsets, padding=grid.extrapolation, stack_dim=stack_dim)
    return [CenteredGrid(data[i], grid.box, grid.extrapolation) for i in range(len(offsets))]


def stagger(field: CenteredGrid, face_function: callable, extrapolation: math.extrapolation.Extrapolation, type: type = StaggeredGrid):
    """
    Creates a new grid by evaluating `face_function` given two neighbouring cells.
    One layer of missing cells is inferred from the extrapolation.
    
    This method returns a Field of type `type` which must be either StaggeredGrid or CenteredGrid.
    When returning a StaggeredGrid, the new values are sampled at the faces of neighbouring cells.
    When returning a CenteredGrid, the new grid has the same resolution as `field`.

    Args:
      field: centered grid
      face_function: function mapping (value1: Tensor, value2: Tensor) -> center_value: Tensor
      extrapolation: extrapolation mode of the returned grid. Has no effect on the values.
      type: one of (StaggeredGrid, CenteredGrid)
      field: CenteredGrid: 
      face_function: callable: 
      extrapolation: math.extrapolation.Extrapolation: 
      type: type:  (Default value = StaggeredGrid)

    Returns:
      grid of type matching the `type` argument

    """
    all_lower = []
    all_upper = []
    if type == StaggeredGrid:
        for dim in field.shape.spatial.names:
            all_upper.append(math.pad(field.values, {dim: (0, 1)}, field.extrapolation))
            all_lower.append(math.pad(field.values, {dim: (1, 0)}, field.extrapolation))
        all_upper = math.channel_stack(all_upper, 'vector')
        all_lower = math.channel_stack(all_lower, 'vector')
        values = face_function(all_lower, all_upper)
        return StaggeredGrid(values, field.bounds, extrapolation)
    elif type == CenteredGrid:
        left, right = math.shift(field.values, (-1, 1), padding=field.extrapolation, stack_dim='vector')
        values = face_function(left, right)
        return CenteredGrid(values, field.bounds, extrapolation)
    else:
        raise ValueError(type)


def divergence(field: Grid) -> CenteredGrid:
    """
    Computes the divergence of a grid using finite differences.

    This function can operate in two modes depending on the type of `field`:

    * `CenteredGrid` approximates the divergence at cell centers using central differences
    * `StaggeredGrid` exactly computes the divergence at cell centers

    Args:
        field: vector field as `CenteredGrid` or `StaggeredGrid`

    Returns:
        Divergence field as `CenteredGrid`
    """
    if isinstance(field, StaggeredGrid):
        components = []
        for i, dim in enumerate(field.shape.spatial.names):
            div_dim = math.gradient(field.values.vector[i], dx=field.dx[i], difference='forward', padding=None, dims=[dim]).gradient[0]
            components.append(div_dim)
        data = math.sum(components, 0)
        return CenteredGrid(data, field.box, field.extrapolation.gradient())
    elif isinstance(field, CenteredGrid):
        left, right = shift(field, (-1, 1), stack_dim='div_')
        grad = (right - left) / (field.dx * 2)
        components = [grad.vector[i].div_[i] for i in range(grad.div_.size)]
        result = sum(components)
        return result
    else:
        raise NotImplementedError(f"{type(field)} not supported. Only StaggeredGrid allowed.")


FieldType = TypeVar('FieldType', bound=Field)
GridType = TypeVar('GridType', bound=Grid)


def minimize(function, x0: Grid, solve_params: math.Solve):
    data_function = _operate_on_values(function, x0)
    converged, x, iterations = math.minimize(data_function, x0.values, solve_params=solve_params)
    return converged, x0.with_(values=x), iterations


def solve(function, y: Grid, x0: Grid, solve_params: math.Solve, constants: tuple or list = (), callback=None):
    if callback is not None:
        def field_callback(x):
            x = x0.with_(values=x)
            callback(x)
    else:
        field_callback = None
    data_function = _operate_on_values(function, x0)
    constants = [c.values if isinstance(c, SampledField) else c for c in constants]
    assert all(isinstance(c, math.Tensor) for c in constants)
    converged, x, iterations = math.solve(data_function, y.values, x0.values, solve_params=solve_params, constants=constants, callback=field_callback)
    return converged, x0.with_(values=x), iterations


def _operate_on_values(field_function, *proto_fields):
    """
    Constructs a wrapper function operating on field values from a function operating on fields.
    The wrapper function assembles fields and calls `field_function`.

    This is useful when passing functions to a `phi.math` operation, e.g. `phi.math.solve()`.

    Args:
        field_function: Function whose arguments are fields
        *proto_fields: To specify non-value properties of the fields.

    Returns:
        Wrapper for `field_function` that takes the field values of as input and returns the field values of the result.
    """
    @wraps(field_function)
    def wrapper(*field_data):
        fields = [proto.with_(values=data) for data, proto in zip(field_data, proto_fields)]
        result = field_function(*fields)
        if isinstance(result, math.Tensor):
            return result
        elif isinstance(result, SampledField):
            return result.values
        else:
            raise ValueError(f"function must return an instance of SampledField or Tensor but returned {result}")
    return wrapper


def data_bounds(field: SampledField):
    data = field.points
    min_vec = math.min(data, dim=data.shape.spatial.names)
    max_vec = math.max(data, dim=data.shape.spatial.names)
    return Box(min_vec, max_vec)


def mean(field: Grid):
    return math.mean(field.values, field.shape.spatial)


def normalize(field: SampledField, norm: SampledField, epsilon=1e-5):
    data = math.normalize_to(field.values, norm.values, epsilon)
    return field.with_(values=data)


def pad(grid: Grid, widths: int or tuple or list or dict):
    if isinstance(widths, int):
        widths = {axis: (widths, widths) for axis in grid.shape.spatial.names}
    elif isinstance(widths, (tuple, list)):
        widths = {axis: (width if isinstance(width, (tuple, list)) else (width, width)) for axis, width in zip(grid.shape.spatial.names, widths)}
    else:
        assert isinstance(widths, dict)
    widths_list = [widths[axis] for axis in grid.shape.spatial.names]
    if isinstance(grid, Grid):
        data = math.pad(grid.values, widths, grid.extrapolation)
        w_lower = math.tensor([w[0] for w in widths_list])
        w_upper = math.tensor([w[1] for w in widths_list])
        box = Box(grid.box.lower - w_lower * grid.dx, grid.box.upper + w_upper * grid.dx)
        return type(grid)(data, box, grid.extrapolation)
    raise NotImplementedError(f"{type(grid)} not supported. Only Grid instances allowed.")


def concat(*fields: SampledField, dim: str):
    assert all(isinstance(f, SampledField) for f in fields)
    assert all(isinstance(f, type(fields[0])) for f in fields)
    if any(f.extrapolation != fields[0].extrapolation for f in fields):
        raise NotImplementedError("Concatenating extrapolations not supported")
    if isinstance(fields[0], Grid):
        values = math.concat([f.values for f in fields], dim=dim)
        return fields[0].with_(values=values)
    elif isinstance(fields[0], PointCloud):
        elements = geom.concat([f.elements for f in fields], dim, sizes=[f.shape.get_size(dim) for f in fields])
        values = math.concat([math.expand(f.values, dim, f.shape.get_size(dim)) for f in fields], dim)
        colors = math.concat([math.expand(f.color, dim, f.shape.get_size(dim)) for f in fields], dim)
        return fields[0].with_(elements=elements, values=values, color=colors)
    raise NotImplementedError(type(fields[0]))


def batch_stack(*fields, dim: str):
    assert all(isinstance(f, SampledField) for f in fields)
    assert all(isinstance(f, type(fields[0])) for f in fields)
    if any(f.extrapolation != fields[0].extrapolation for f in fields):
        raise NotImplementedError("Concatenating extrapolations not supported")
    if isinstance(fields[0], Grid):
        values = math.batch_stack([f.values for f in fields], dim)
        return fields[0].with_(values=values)
    elif isinstance(fields[0], PointCloud):
        elements = geom.stack(*[f.elements for f in fields], dim=dim)
        values = math.batch_stack([f.values for f in fields], dim=dim)
        colors = math.batch_stack([f.color for f in fields], dim=dim)
        return fields[0].with_(elements=elements, values=values, color=colors)
    raise NotImplementedError(type(fields[0]))


def real(field: Field):
    if isinstance(field, SampledField):
        return field.with_(values=math.real(field.values))
    raise NotImplementedError()


def imag(field: Field):
    if isinstance(field, SampledField):
        return field.with_(values=math.imag(field.values))
    raise NotImplementedError()


def assert_close(*fields: SampledField,
                 rel_tolerance: float = 1e-5,
                 abs_tolerance: float = 0):
    f0 = next(filter(lambda t: isinstance(t, SampledField), fields))
    values = [(f >> f0).values if isinstance(f, SampledField) else math.tensor(f) for f in fields]
    math.assert_close(*values, rel_tolerance=rel_tolerance, abs_tolerance=abs_tolerance)


# def staggered_curl_2d(grid, pad_width=(1, 2)):
#     assert isinstance(grid, CenteredGrid)
#     kernel = math.zeros((3, 3, 1, 2))
#     kernel[1, :, 0, 0] = [0, 1, -1]  # y-component: - dz/dx
#     kernel[:, 1, 0, 1] = [0, -1, 1]  # x-component: dz/dy
#     scalar_potential = grid.padded([pad_width, pad_width]).values
#     vector_field = math.conv(scalar_potential, kernel, padding='valid')
#     return StaggeredGrid(vector_field, bounds=grid.box)


def where(mask: Field or Geometry, field_true: Field, field_false: Field):
    if isinstance(mask, Geometry):
        mask = HardGeometryMask(mask)
    elif isinstance(mask, SampledField):
        field_true = field_true.at(mask)
        field_false = field_false.at(mask)
    elif isinstance(field_true, SampledField):
        mask = mask.at(field_true)
        field_false = field_false.at(field_true)
    elif isinstance(field_false, SampledField):
        mask = mask.at(field_true)
        field_true = field_true.at(mask)
    else:
        raise NotImplementedError('At least one argument must be a SampledField')
    values = mask.values * field_true.values + (1 - mask.values) * field_false.values
    # values = math.where(mask.values, field_true.values, field_false.values)
    return field_true.with_(values=values)


def l2_loss(field: SampledField, batch_norm=True):
    """ L2 loss for the unweighted values of the field. See `phi.math.l2_loss()`. """
    return math.l2_loss(field.values, batch_norm=batch_norm)


def stop_gradient(field: SampledField):
    """ See `phi.math.stop_gradient()` """
    if isinstance(field, Grid):
        return field.with_(values=math.stop_gradient(field.values))
    raise NotImplementedError()


def extrapolate_valid(grid: GridType, valid: GridType, distance_cells=1) -> tuple:
    """
    Extrapolates values of `grid` which are marked by nonzero values in `valid` using `phi.math.extrapolate_valid_values().
    If `values` is a StaggeredGrid, its components get extrapolated independently.

    Args:
        grid: Grid holding the values for extrapolation
        valid: Grid (same type as `values`) marking the positions for extrapolation with nonzero values
        distance_cells: Number of extrapolation steps

    Returns:
        grid: Grid with extrapolated values.
        valid: binary Grid marking all valid values after extrapolation.
    """
    assert isinstance(valid, type(grid)), 'Type of valid Grid must match type of grid.'
    if isinstance(grid, CenteredGrid):
        new_values, new_valid = extrapolate_valid_values(grid.values, valid.values, distance_cells)
        return grid.with_(values=new_values), valid.with_(values=new_valid)
    elif isinstance(grid, StaggeredGrid):
        new_values = []
        new_valid = []
        for cgrid, cvalid in zip(grid.unstack('vector'), valid.unstack('vector')):
            new_tensor, new_mask = extrapolate_valid(cgrid, valid=cvalid, distance_cells=distance_cells)
            new_values.append(new_tensor.values)
            new_valid.append(new_mask.values)
        return grid.with_(values=math.channel_stack(new_values, 'vector')), valid.with_(values=math.channel_stack(new_valid, 'vector'))
    else:
        raise NotImplementedError()





def assert_minimal_config():  # raises AssertionError
    import sys
    assert sys.version_info.major == 3 and sys.version_info.minor >= 6, "phiflow requires Python 3.6 or newer to run"

    try:
        import numpy
    except ImportError:
        raise AssertionError("phiflow is unable to run because NumPy is not installed.")
    try:
        import scipy
    except ImportError:
        raise AssertionError("phiflow is unable to run because SciPy is not installed.")
    from phi import flow
    from phi import math
    math.assert_close(math.ones(batch=8, x=64) + math.ones(batch=8, x=64), 2)


def troubleshoot():
    import phi
    return f"PhiFlow version {phi.__version__}\n"\
           f"Web interface: {troubleshoot_dash()}\n"\
           f"TensorFlow: {troubleshoot_tensorflow()}\n"\
           f"PyTorch: {troubleshoot_torch()}\n"\
           f"Jax: {troubleshoot_jax()}"


def troubleshoot_tensorflow():
    from phi import math
    import os
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # only errors
    try:
        import tensorflow
    except ImportError:
        return "Not installed."
    try:
        from phi import tf
    except BaseException as err:
        return f"Installed but not available due to internal error: {err}"
    try:
        gpu_count = len(tf.TF_BACKEND.list_devices('GPU'))
    except BaseException as err:
        return f"Installed but device initialization failed with error: {err}"
    with tf.TF_BACKEND:
        try:
            math.assert_close(math.ones(batch=8, x=64) + math.ones(batch=8, x=64), 2)
            # TODO cuDNN math.conv(math.ones(batch=8, x=64), math.ones(x=4))
        except BaseException as err:
            return f"Installed but tests failed with error: {err}"
    if gpu_count == 0:
        return f"Installed, {gpu_count} GPUs available."
    else:
        from phi.tf._tf_cuda_resample import librariesLoaded
        if librariesLoaded:
            cuda_str = 'CUDA kernels available.'
        else:
            import platform
            if platform.system().lower() != 'linux':
                cuda_str = f"CUDA kernels not available and compilation not recommended on {platform.system()}."
            else:
                cuda_str = f"CUDA kernels not available. Clone the phiflow source from GitHub and run 'python setup.py tf_cuda' to compile them."
        return f"Installed, {gpu_count} GPUs available. {cuda_str}"


def troubleshoot_torch():
    from phi import math
    try:
        import torch
    except ImportError:
        return "Not installed."
    try:
        from phi import torch
    except BaseException as err:
        return f"Installed but not available due to internal error: {err}"
    try:
        gpu_count = len(torch.TORCH_BACKEND.list_devices('GPU'))
    except BaseException as err:
        return f"Installed but device initialization failed with error: {err}"
    with torch.TORCH_BACKEND:
        try:
            math.assert_close(math.ones(batch=8, x=64) + math.ones(batch=8, x=64), 2)
        except BaseException as err:
            return f"Installed but tests failed with error: {err}"
    return f"Installed, {gpu_count} GPUs available."


def troubleshoot_jax():
    from phi import math
    try:
        import jax
        import jaxlib
    except ImportError:
        return "Not installed."
    try:
        from phi import jax
    except BaseException as err:
        return f"Installed but not available due to internal error: {err}"
    try:
        gpu_count = len(jax.JAX_BACKEND.list_devices('GPU'))
    except BaseException as err:
        return f"Installed but device initialization failed with error: {err}"
    with jax.JAX_BACKEND:
        try:
            math.assert_close(math.ones(batch=8, x=64) + math.ones(batch=8, x=64), 2)
        except BaseException as err:
            return f"Installed but tests failed with error: {err}"
    return f"Installed, {gpu_count} GPUs available."


def troubleshoot_dash():
    try:
        import dash
    except ImportError:
        return "Dash not installed"
    try:
        import plotly
    except ImportError:
        return "Plotly not installed"
    try:
        import imageio
    except ImportError:
        return "ImageIO not installed"
    try:
        import matplotlib
    except ImportError:
        return "Matplotlib not installed"
    try:
        dash.Dash('Test')
    except BaseException as e:
        return f"Runtime error: {e}"
    return 'OK'




import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output

from . import viewsettings, model_controls
from .dash_plotting import dash_graph_plot, EMPTY_FIGURE
from .player_controls import STEP_COMPLETE, REFRESH_INTERVAL
import webglviewer
from .viewsettings import parse_view_settings
from .webgl_util import default_sky, EMPTY_GRID, webgl_prepare_data

# --- Viewer ---

def build_viewer(dashapp, height, initial_field_name=None, id='viewer', config=None):

    field_options = [{'label': item, 'value': item} for item in dashapp.app.fieldnames] + [{'label': '<None>', 'value': 'None'}]
    if initial_field_name is None:
        initial_field_name = dashapp.app.fieldnames[0] if len(dashapp.app.fieldnames) > 0 else 'None'

    layout = html.Div(style={'height': '100%'}, children=[
        html.Div(style={'width': '100%', 'height': '5%', 'display': 'inline-block', 'vertical-align': 'middle'}, children=[
            html.Div(style={'width': '90%', 'display': 'inline-block'}, children=[
                dcc.Dropdown(options=field_options, value=initial_field_name, id=id+'-field-select'),
            ]),
            html.Div(style={'width': '10%', 'display': 'inline-block', 'vertical-align': 'top', 'text-align': 'center'}, children=[
                dcc.Checklist(options=[{'label': '3D', 'value': '3D'}], value=[], id=id+'-3d', style={'margin-top': 6}),
            ]),
        ]),
        html.Div(id=id+'-figure-container', children=[], style={'height': '95%', 'width': '100%', 'display': 'inline-block'}),
    ])

    @dashapp.dash.callback(Output(id+'-figure-container', 'children'), [Input(id+'-3d', 'value')])
    def choose_viewer(list3d):
        if list3d:
            return [
                dcc.Interval(id=id + '-webgl-initializer', interval=100, max_intervals=1),
                webglviewer.Webglviewer(id=id+'-webgl', sky=default_sky(), material_type='LIGHT_SMOKE', representation_type='DENSITY')
            ]
        else:
            return dcc.Graph(figure=EMPTY_FIGURE, id=id + '-graph', style={'height': '100%'})

    @dashapp.dash.callback(Output(id+'-graph', 'figure'), (Input(id+'-field-select', 'value'), STEP_COMPLETE, REFRESH_INTERVAL) + viewsettings.VIEW_SETTINGS + tuple(model_controls.MODEL_CONTROLS + model_controls.MODEL_ACTIONS))
    def update_figure(field, _0, _1, *settings):
        if field is None or field == 'None':
            return EMPTY_FIGURE
        data = dashapp.get_field(field)
        if data is None:
            return EMPTY_FIGURE
        settings_dict = parse_view_settings(config, *settings)
        settings_dict['minmax'] = dashapp.get_minmax(field)
        settings_dict['figsize'] = height
        return dash_graph_plot(data, settings_dict)

    @dashapp.dash.callback(Output(id+'-webgl', 'data'), (Input(id+'-field-select', 'value'), Input(id+'-webgl-initializer', 'n_intervals'), STEP_COMPLETE, REFRESH_INTERVAL) + viewsettings.VIEW_SETTINGS + tuple(model_controls.MODEL_CONTROLS + model_controls.MODEL_ACTIONS))
    def update_data(field, _0, _1, _2, *settings):
        if field is None or field == 'None':
            return EMPTY_GRID
        data = dashapp.get_field(field)
        if data is None:
            return EMPTY_GRID
        settings_dict = parse_view_settings(config, *settings)
        return webgl_prepare_data(data, settings_dict)

    return layout



# pylint: disable-msg = wildcard-import, unused-wildcard-import, unused-import
"""
Standard import for PyTorch mode.

Extends the import `from phi.flow import *` by PyTorch-related functions and modules.

The following TensorFlow modules are included: `torch`, `torchf` (short for `torch.nn.functional`).

Importing this module registers the PyTorch backend as the default backend.
New tensors created via `phi.math` functions will be backed by PyTorch tensors.

See `phi.flow`, `phi.tf.flow`.
"""

from phi.flow import *
from ._torch_util import variable
from ._torch_backend import TORCH_BACKEND
import torch
import torch.nn.functional as torchf

backend.set_global_default_backend(TORCH_BACKEND)



"""
Worlds are used to manage simulations consisting of multiple states.
A world uses a StateCollection with CollectivePhysics to resolve state dependencies.
Worlds also facilitate referencing states when performing a forward simulation.

A default World, called `world` is provided for convenience.
"""
import inspect
import warnings
from typing import TypeVar

from phi import geom
from phi.field import GeometryMask
from phi.physics._effect import Gravity
from ._physics import Physics, State, struct, _as_physics, Static


class StateProxy(object):
    """
    StateProxy mirrors all data_dict of a state in an associated world.
    While State objects are generally immutable, StateProxy also implements setting any attribute of the state.
    When an attribute is set, a copy of the state with the new value replaces the old state in the world.
    This object then mirrors the values of the new state.
    
    After world.step() is invoked, all Proxies of that world will mirror the state after stepping.
    
    To reference the current immutable state of

    Args:

    Returns:

    """

    def __init__(self, enclosing_world, state_name):
        self.world = enclosing_world
        self.state_name = state_name

    @property
    def state(self):
        """
        Finds and returns the state in the referenced world that matches the StateProxy's state_name.
        :return: State

        Args:

        Returns:

        """
        state = self.world.state[self.state_name]
        assert state is not None
        return state

    @state.setter
    def state(self, state):
        """
        Replaces the State in the referenced world.

        Args:
          state: 

        Returns:

        """
        assert state.name == self.state_name
        self.world.state = self.world.state.state_replaced(state)

    @property
    def physics(self):
        """
        Returns the Physics object used by the referenced world for the system.
        If not specified manually, the default physics is returned.

        Args:

        Returns:

        """
        physics = self.world.physics.for_(self.state)
        assert physics is not None
        return physics

    @physics.setter
    def physics(self, physics):
        """
        Sets a specific Physics object for the system in the referenced world.

        Args:
          physics: Physics

        Returns:

        """
        assert isinstance(physics, Physics)
        self.world.physics.add(self.state_name, physics)

    def step(self, dt=1.0, physics=None):
        """
        Steps this system in the referenced world. Other states are not affected.

        Args:
          dt: time increment (Default value = 1.0)
          physics: specific Physics object to use, defaults to self.physics

        Returns:

        """
        self.world.step(self, dt=dt, physics=physics)

    def __getattr__(self, item):
        """
Returns an attribute from the referenced State.
        :param item: attribute name
        """
        assert item not in ('world', 'state_name', 'physics', 'state')
        return getattr(self.state, item)

    def __setattr__(self, key, value):
        """
Changes the referenced state by replacing it in the referenced world.
        :param key: State attribute name
        :param value: new value
        """
        if key in ('world', 'state_name', 'physics', 'state'):
            object.__setattr__(self, key, value)
        else:
            self.state = self.state.copied_with(**{key:value})


# pylint: disable-msg = invalid-name
S = TypeVar('S', bound=State)


class World(object):
    """
    A world object defines a global state as well as a set of rules (Physics objects) that definition how the state evolves.
    
    The world manages dependencies among the contained simulations and provides convenience methods for creating proxies for specific simulations.
    
    The method world.step() evolves the whole state or optionally a specific state in time.

    Args:

    Returns:

    """

    def __init__(self, batch_size=None, add_default_objects=True):
        # --- Insert object / create proxy shortcuts ---
        self._state = self.physics = self.observers = self.batch_size = None
        self.reset(batch_size, add_default_objects)

    def reset(self, batch_size=None, add_default_objects=True):
        """
        Resets the world to the default configuration.
        This removes all States and observers.

        Args:
          batch_size: int or None (Default value = None)
          add_default_objects: if True, adds defaults like Gravity

        Returns:

        """
        self._state = StateCollection()
        self.physics = self._state.default_physics()
        self.observers = set()
        self.batch_size = batch_size
        if add_default_objects:
            self.add(Gravity())

    @property
    def state(self):
        """
        Returns the current state of the world.
        :return: StateCollection

        Args:

        Returns:

        """
        return self._state

    @property
    def age(self):
        """Alias for world.state.age"""
        return self._state.age

    @state.setter
    def state(self, state):
        """
        Sets the current state of the world and informs all observers.

        Args:
          state: StateCollection

        Returns:

        """
        assert state is not None
        assert isinstance(state, StateCollection)
        self._state = state
        for observer in self.observers:
            observer(self)

    def step(self, state=None, dt=1.0, physics=None):
        """
        Evolves the current world state by a time increment dt.
        If state is provided, only that state is evolved, leaving the others untouched.
        The optional physics parameter can then be used to override the default physics.
        Otherwise, all states are evolved.
        
        Calling World.step resolves all dependencies among simulations and then calls Physics.step on each simulation to evolve the states.
        
        Invoking this method alters the world state. To to_field a copy of the state, use :func:`World.stepped <~world.World.stepped>` instead.

        Args:
          state: State, StateProxy or None (Default value = None)
          dt: time increment (Default value = 1.0)
          physics: Physics object for the state or None for default

        Returns:
          evolved state if a specific state was provided

        """
        if state is None:
            if physics is None:
                physics = self.physics
            self.state = physics.step(self._state, dt)
            return self.state
        else:
            if isinstance(state, StateProxy):
                state = state.state
            s = self.physics.substep(state, self._state, dt, override_physics=physics)
            self.state = self._state.state_replaced(s)
            return s

    def stepped(self, state=None, dt=1.0, physics=None):
        """
        Similar to step() but does not change the state of the world. Instead, the new state is returned.

        Args:
          state:  (Default value = None)
          dt:  (Default value = 1.0)
          physics:  (Default value = None)

        Returns:

        """
        if state is None:
            if physics is None:
                physics = self.physics
            return physics.step(self._state, None, dt)
        else:
            if isinstance(state, StateProxy):
                state = state.state
            return self.physics.substep(state, self._state, dt, override_physics=physics)

    def add(self, state, physics=None):
        # type: (S, Physics) -> S
        """
Adds a State to the world that will be stepped forward in time each time world.step() is invoked.
        :param state: State or list of States
        :param physics: (optional) Physics to use during world.step(). If a list was provided for `state`, a matching list must be given for `state`.
        :return: StateProxy referencing the current state of the added system. If world.state is updated (e.g. because world.step() was called), the StateProxy will refer to the updated values.
        """
        if isinstance(state, dict):
            raise ValueError('Cannot add dict to world. Maybe you meant world.add(**dict)?')
        if isinstance(state, (tuple, list)):
            assert isinstance(physics, (tuple, list))
            assert len(state) == len(physics)
            return [self.add(s, p) for s, p in zip(state, physics)]
        else:
            if physics is not None:
                self.physics.add(state.name, physics)
            elif state.default_physics() is not None and not isinstance(state.default_physics(), Static):
                warnings.warn('No physics provided to world.add(%s). In the future this will default to static physics' % state)
            self.state = self.state.state_added(state)
            return StateProxy(self, state.name)

    def add_all(self, *states):
        """
        Add a collection of states to the system using world.add(state).

        Args:
          *states: 

        Returns:

        """
        warnings.warn('World.add_all() is deprecated. Use World.add(list_of_states) instead.', DeprecationWarning)
        for state in states:
            self.add(state)

    def remove(self, obj):
        """
        Remove a system or collection of systems from the world.

        Args:
          obj: one of the following: State, state name, subclass of State, tuple or list thereof

        Returns:

        """
        if inspect.isclass(obj):
            states = self.state.all_instances(obj)
            self.remove(states)
        elif isinstance(obj, (tuple, list)):
            for state in obj:
                self.remove(state)
        else:
            key = obj if isinstance(obj, str) else obj.name
            self.state = self.state.state_removed(key)
            self.physics.remove(key)

    def get_physics(self, state):
        """
        Looks up the Physics object associated with a given State or StateProxy.
        If no Physics object was registered manually, the state.default_physics() object is used.

        Args:
          state: State or StateProxy contained in this world

        Returns:
          Physics

        """
        if isinstance(state, StateProxy):
            state = state.state
        return self.physics.for_(state)

    def __getattr__(self, item):
        if item in self.state:
            return StateProxy(self, item)
        else:
            return object.__getattribute__(self, item)


def obstacle_mask(world_or_proxy):
    """
    Builds a binary Field, masking all obstacles in the world.

    Args:
      world_or_proxy: World or StateProxy object

    Returns:
      Field

    """
    world = world_or_proxy.world if isinstance(world_or_proxy, StateProxy) else world_or_proxy
    assert isinstance(world, World)
    geometries = [obstacle.geometry for obstacle in world.state.all_with_tag('obstacle')]
    return GeometryMask(geom.union(*geometries))


class StateCollection(dict):

    def __init__(self, states=None):
        """
        Create a state collection from a dictionary of states.

        Args:
          states(dict or list or tuple None): dict mapping from state names to states
        """
        if states is None:
            states = {}
        elif not isinstance(states, dict):
            states = {state.name: state for state in states}
        dict.__init__(self, states)

    def __setitem__(self, key, val):
        raise AttributeError('StateCollections are immutable')

    def all_with_tag(self, tag):
        return [s for s in self.values() if tag in s.tags]

    def all_instances(self, cls):
        return [s for s in self.values() if isinstance(s, cls)]

    def state_added(self, state):
        assert state.name not in self, 'A state with name "%s" is already present. Use state_replaced() to replace it.' % state.name
        new_states = self.copy()
        new_states[state.name] = state
        return StateCollection(new_states)

    def state_replaced(self, new_state):
        assert new_state.name in self, 'No state found with name "%s"' % new_state.name
        new_states = dict(self)
        new_states[new_state.name] = new_state
        return StateCollection(new_states)

    def state_removed(self, state):
        name = state if isinstance(state, str) else state.name
        new_states = dict(self)
        del new_states[name]
        return StateCollection(new_states)

    def find(self, name):
        warnings.warn("StateCollection.find is deprecated. Use statecollection[name] instead.", DeprecationWarning)
        return dict.__getitem__(self, name)

    def __getitem__(self, item):
        if isinstance(item, State):
            return self[item.name]
        if isinstance(item, str):
            return dict.__getitem__(self, item)
        if struct.isstruct(item):
            return struct.map(lambda x: self[x], item, content_type=struct.INVALID)
        try:
            return self[item.name]
        except AttributeError as e:
            pass
        raise ValueError('Illegal argument: %s' % item)

    def __getattr__(self, item):
        return self[item]

    def default_physics(self):
        warnings.warn("StateCollection will be removed in the future.", DeprecationWarning)
        return CollectivePhysics()

    def __repr__(self):
        return '[' + ', '.join((str(s) for s in self)) + ']'

    def __contains__(self, item):
        if isinstance(item, State):
            return item.name in self
        if isinstance(item, str):
            return dict.__contains__(self, item)
        raise ValueError('Illegal type: %s' % type(item))

    def __hash__(self):
        return 0

    @property
    def states(self):
        return self

    def copied_with(self, **kwargs):
        if len(kwargs) == 0:
            return self
        assert len(kwargs) == 1
        name, value = next(iter(kwargs.items()))
        assert name == 'states'
        return StateCollection(value)

    @property
    def shape(self):
        return StateCollection({name: state.shape for name, state in self.items()})

    @property
    def staticshape(self):
        return StateCollection({name: state.staticshape for name, state in self.items()})

    @property
    def dtype(self):
        return StateCollection({name: state.dtype for name, state in self.items()})


CollectiveState = StateCollection


class CollectivePhysics(Physics):

    def __init__(self):
        Physics.__init__(self, {})
        self.physics = {}  # map from name to Physics

    def step(self, state_collection, dt=1.0, **dependent_states):
        assert len(dependent_states) == 0
        if len(state_collection) == 0:
            return state_collection
        unhandled_states = list(state_collection.values())
        next_states = {}
        partial_next_state_collection = StateCollection(next_states)

        for sweep in range(len(state_collection)):
            for state in tuple(unhandled_states):
                physics = self.for_(state)
                if self._all_dependencies_fulfilled(physics.blocking_dependencies, state_collection, partial_next_state_collection):
                    next_state = self.substep(state, state_collection, dt, partial_next_state_collection=partial_next_state_collection)
                    assert next_state is not None, "step() called on %s returned None for state '%s'" % (type(physics).__name__, state)
                    assert isinstance(next_state, State), "step() called on %s dit not return a State but '%s' for state '%s'" % (type(physics).__name__, next_state, state)
                    assert next_state.name == state.name, "The state name must remain constant during step(). Caused by '%s' on state '%s'." % (type(physics).__name__, state)
                    next_states[next_state.name] = next_state
                    unhandled_states.remove(state)
            partial_next_state_collection = StateCollection(next_states)
            if len(unhandled_states) == 0:
                ordered_states = [partial_next_state_collection[state] for state in state_collection]
                return StateCollection(ordered_states)

        # Error
        errstr = 'Cyclic blocking_dependencies in simulation: %s' % unhandled_states
        for state in tuple(unhandled_states):
            physics = self.for_(state)
            state_dict = self._gather_dependencies(physics.blocking_dependencies, state_collection, {})
            errstr += '\nState "%s" with physics "%s" depends on %s' % (state, physics, state_dict)
        raise AssertionError(errstr)

    def substep(self, state, state_collection, dt, override_physics=None, partial_next_state_collection=None):
        physics = self.for_(state) if override_physics is None else override_physics
        # --- gather dependencies
        dependent_states = {}
        self._gather_dependencies(physics.dependencies, state_collection, dependent_states)
        if partial_next_state_collection is not None:
            self._gather_dependencies(physics.blocking_dependencies, partial_next_state_collection, dependent_states)
        # --- execute step ---
        next_state = physics.step(state, dt, **dependent_states)
        return next_state

    def _gather_dependencies(self, dependencies, state_collection, result_dict):
        for statedependency in dependencies:
            if statedependency.state_name is not None:
                matching_states = state_collection.find(statedependency.state_name)
            else:
                matching_states = state_collection.all_with_tag(statedependency.tag)
            if statedependency.single_state:
                assert len(matching_states) == 1, 'Dependency %s requires 1 state but found %d' % (statedependency, len(matching_states))
                value = matching_states[0]
            else:
                value = tuple(matching_states)
            result_dict[statedependency.parameter_name] = value
        return result_dict

    def _all_dependencies_fulfilled(self, dependencies, all_states, computed_states):
        state_dict = self._gather_dependencies(dependencies, all_states, {})
        for name, states in state_dict.items():
            if isinstance(states, tuple):
                for state in states:
                    if state.name not in computed_states:
                        return False
            else:  # single state
                if states.name not in computed_states:
                    return False
        return True

    def for_(self, state):
        return self.physics[state.name] if state.name in self.physics else state.default_physics()

    def add(self, name, physics):
        self.physics[name] = _as_physics(physics)

    def remove(self, name):
        if name in self.physics:
            del self.physics[name]


world = World()



"""
The fields module provides a number of data structures and functions to represent continuous, spatially varying data.

All fields are subclasses of Field which provides abstract functions for sampling field values at physical locations.

The most commonly used field types are

* CenteredGrid embeds a tensor in the physical space. Uses linear interpolation between grid points.
* StaggeredGrid samples the vector components at face centers instead of at cell centers.
* Noise is a function that produces a procedurally generated noise field

See the `phi.field` module documentation at https://tum-pbs.github.io/PhiFlow/Fields.html
"""

from ._field import Field, SampledField
from ._analytic import AnalyticField
from ._constant import ConstantField
from ._mask import HardGeometryMask, SoftGeometryMask as GeometryMask, SoftGeometryMask
from ._grid import Grid, CenteredGrid, StaggeredGrid, unstack_staggered_tensor, stack_staggered_components
from ._point_cloud import PointCloud
from ._noise import Noise
from ._angular_velocity import AngularVelocity
from ._field_math import (
    assert_close,
    laplace, gradient, divergence, stagger,
    mean, pad, shift, normalize,
    concat, batch_stack,
    real, imag,
    solve, minimize,
    where,
    l2_loss,
    stop_gradient,
    extrapolate_valid,
)
from ._field_io import write, read
from ._scene import Scene

__all__ = [key for key in globals().keys() if not key.startswith('_')]



# Because division is different in Python 2 and 3
from __future__ import division

import numpy as np
from typing import Tuple
from phi import struct
from . import extrapolation as extrapolation
from . import _functions as math
from ._config import GLOBAL_AXIS_ORDER
from .extrapolation import Extrapolation
from ._functions import channel_stack
from ._shape import Shape
from ._tensors import Tensor
from ._tensors import tensor


def spatial_sum(value: Tensor):
    return math.sum_(value, dim=value.shape.spatial.names)


def vec_abs(vec: Tensor):
    return math.sqrt(math.sum_(vec ** 2, dim=vec.shape.channel.names))


def vec_squared(vec: Tensor):
    return math.sum_(vec ** 2, dim=vec.shape.channel.names)


def cross_product(vec1: Tensor, vec2: Tensor):
    vec1, vec2 = math.tensors(vec1, vec2)
    spatial_rank = vec1.vector.size if 'vector' in vec1.shape else vec2.vector.size
    if spatial_rank == 2:  # Curl in 2D
        dist_0, dist_1 = vec2.vector.unstack()
        if GLOBAL_AXIS_ORDER.is_x_first:
            velocity = vec1 * math.channel_stack([-dist_1, dist_0], 'vector')
        else:
            velocity = vec1 * math.channel_stack([dist_1, -dist_0], 'vector')
        return velocity
    elif spatial_rank == 3:  # Curl in 3D
        raise NotImplementedError(f'spatial_rank={spatial_rank} not yet implemented')
    else:
        raise AssertionError(f'dims = {spatial_rank}. Vector product not available in > 3 dimensions')


def normalize_to(target: Tensor, source: Tensor, epsilon=1e-5):
    """
    Multiplies the target so that its total content matches the source.

    Args:
      target: a tensor
      source: a tensor or number
      epsilon: small number to prevent division by zero or None. (Default value = 1e-5)
      target: Tensor: 
      source: Tensor: 

    Returns:
      normalized tensor of the same shape as target

    """
    target_total = math.sum_(target, dim=target.shape.non_batch.names)
    denominator = math.maximum(target_total, epsilon) if epsilon is not None else target_total
    source_total = math.sum_(source, dim=source.shape.non_batch.names)
    return target * (source_total / denominator)


def l1_loss(tensor: Tensor, batch_norm=True, reduce_batches=True):
    """
    get L1 loss

    Args:
      tensor: Tensor: 
      batch_norm:  (Default value = True)
      reduce_batches:  (Default value = True)

    Returns:

    """
    if struct.isstruct(tensor):
        all_tensors = struct.flatten(tensor)
        return sum(l1_loss(tensor, batch_norm, reduce_batches) for tensor in all_tensors)
    if reduce_batches:
        total_loss = math.sum_(math.abs(tensor))
    else:
        total_loss = math.sum_(math.abs(tensor), dim=list(range(1, len(tensor.shape))))
    if batch_norm and reduce_batches:
        batch_size = tensor.shape.sizes[0]
        return math.divide_no_nan(total_loss, math.to_float(batch_size))
    else:
        return total_loss


def l2_loss(tensor: Tensor, batch_norm=True):
    """
    get L2 loss

    Args:
      tensor: Tensor: 
      batch_norm:  (Default value = True)

    Returns:

    """
    return l_n_loss(tensor, 2, batch_norm=batch_norm)


def l_n_loss(tensor: Tensor, n: int, batch_norm=True):
    """
    get Ln loss

    Args:
      tensor: Tensor: 
      n: int: 
      batch_norm:  (Default value = True)

    Returns:

    """
    assert isinstance(tensor, Tensor), f"Must be a Tensor but got {type(tensor).__name__}"
    total_loss = math.sum_(tensor ** n) / n
    if batch_norm:
        batch_size = tensor.shape.batch.volume
        return math.divide_no_nan(total_loss, batch_size)
    else:
        return total_loss


def frequency_loss(tensor, frequency_falloff=100, reduce_batches=True):
    """
    Instead of minimizing each entry of the tensor, minimize the frequencies of the tensor, emphasizing lower frequencies over higher ones.

    Args:
      reduce_batches: whether to reduce the batch dimension of the loss by adding the losses along the first dimension (Default value = True)
      tensor: typically actual - target
      frequency_falloff: large values put more emphasis on lower frequencies, 1.0 weights all frequencies equally. (Default value = 100)

    Returns:
      scalar loss value

    """
    if struct.isstruct(tensor):
        all_tensors = struct.flatten(tensor)
        return sum(frequency_loss(tensor, frequency_falloff, reduce_batches) for tensor in all_tensors)
    diff_fft = abs_square(math.fft(tensor))
    k_squared = math.sum_(math.fftfreq(tensor.shape[1:-1]) ** 2, 'vector')
    weights = math.exp(-0.5 * k_squared * frequency_falloff ** 2)
    return l1_loss(diff_fft * weights, reduce_batches=reduce_batches)


def abs_square(complex):
    """
    get the square magnitude

    Args:
      complex(Tensor): complex input data

    Returns:
      Tensor: real valued magnitude squared

    """
    return math.imag(complex) ** 2 + math.real(complex) ** 2


# Divergence

# def divergence(tensor, dx=1, difference='central', padding='constant', dimensions=None):
#     """
#     Computes the spatial divergence of a vector channel from finite differences.
#
#     :param tensor: vector field; tensor of shape (batch size, spatial dimensions..., spatial rank)
#     :param dx: distance between adjacent grid points (default 1)
#     :param difference: type of difference, one of ('forward', 'central') (default 'forward')
#     :return: tensor of shape (batch size, spatial dimensions..., 1)
#     """
#     assert difference in ('central', 'forward', 'backward'), difference
#     rank = spatial_rank(tensor)
#     if difference == 'forward':
#         return _divergence_nd(tensor, padding, (0, 1), dims) / dx ** rank  # TODO why dx^rank?
#     elif difference == 'backward':
#         return _divergence_nd(tensor, padding, (-1, 0), dims) / dx ** rank
#     else:
#         return _divergence_nd(tensor, padding, (-1, 1), dims) / (2 * dx) ** rank
#
#
# def _divergence_nd(x_, padding, relative_shifts, dims=None):
#     x = tensor(x_)
#     assert x.shape.channel.rank == 1
#     dims = dims if dims is not None else x.shape.spatial.names
#     x = math.pad(x, {axis: (-relative_shifts[0], relative_shifts[1]) for axis in dims}, mode=padding)
#     components = []
#     for dimension in dims:
#         dim_index_in_spatial = x.shape.spatial.reset_indices().index(dimension)
#         lower, upper = _multi_roll(x, dimension, relative_shifts, diminish_others=(-relative_shifts[0], relative_shifts[1]), names=dims, base_selection={0: rank - dimension - 1})
#         components.append(upper - lower)
#     return math.sum_(components, 0)


def shift(x: Tensor,
          offsets: tuple,
          dims: tuple or None = None,
          padding: Extrapolation or None = extrapolation.BOUNDARY,
          stack_dim: str or None = 'shift') -> list:
    """
    shift Tensor by a fixed offset and abiding by extrapolation

    Args:
      x: Input data
      offsets: Shift size
      dims: Dimensions along which to shift, defaults to None
      padding: padding to be performed at the boundary, defaults to extrapolation.BOUNDARY
      stack_dim: dimensions to be stacked, defaults to 'shift'
      x: Tensor: 
      offsets: tuple: 
      dims: tuple or None:  (Default value = None)
      padding: Extrapolation or None:  (Default value = extrapolation.BOUNDARY)
      stack_dim: str or None:  (Default value = 'shift')

    Returns:
      list: offset_tensor

    """
    if stack_dim is None:
        assert len(dims) == 1
    x = tensor(x)
    dims = dims if dims is not None else x.shape.spatial.names
    pad_lower = max(0, -min(offsets))
    pad_upper = max(0, max(offsets))
    if padding is not None:
        x = math.pad(x, {axis: (pad_lower, pad_upper) for axis in dims}, mode=padding)
    offset_tensors = []
    for offset in offsets:
        components = []
        for dimension in dims:
            slices = {dim: slice(pad_lower + offset, -pad_upper + offset) if dim == dimension else slice(pad_lower, -pad_upper) for dim in dims}
            slices = {dim: slice(sl.start, sl.stop if sl.stop < 0 else None) for dim, sl in slices.items()}  # replace stop=0 by stop=None
            components.append(x[slices])
        offset_tensors.append(channel_stack(components, stack_dim) if stack_dim is not None else components[0])
    return offset_tensors


def extrapolate_valid_values(values: Tensor, valid: Tensor, distance_cells: int = 1) -> Tuple[Tensor, Tensor]:
    """
    Extrapolates the values of `values` which are marked by the nonzero values of `valid` for `distance_cells` steps in all spatial directions.
    Overlapping extrapolated values get averaged.

    Examples (1-step extrapolation), x marks the values for extrapolation:
        200   000    210        004   00x    044        100   000    100
        010 + 0x0 => 111        000 + 000 => 204        000 + 000 => 204
        040   000    010        200   x00    220        204   x0x    234

    Args:
        values: Tensor which holds the values for extrapolation
        valid: Tensor with same size as `x` marking the values for extrapolation with nonzero values
        distance_cells: Number of extrapolation steps

    Returns:
        values: Extrapolation result
        valid: mask marking all valid values after extrapolation
    """
    distance_cells = min(distance_cells, max(values.shape))
    for _ in range(distance_cells):
        valid = math.divide_no_nan(valid, valid)  # ensure binary mask
        values_l, values_r = shift(values * valid, (-1, 1))
        mask_l, mask_r = shift(valid, (-1, 1))
        overlap = math.sum_(mask_l + mask_r, dim='shift')
        extp = math.divide_no_nan(math.sum_(values_l + values_r, dim='shift'), overlap)  # take mean where extrapolated values overlap
        new_valid = valid + overlap
        values = math.where(valid, values, math.where(new_valid, extp, values))  # don't overwrite initial values within the mask / keep values not affected by extrapolation
        valid = new_valid
    return values, math.divide_no_nan(valid, valid)


# Gradient

def gradient(grid: Tensor,
             dx: float or int = 1,
             difference: str = 'central',
             padding: Extrapolation or None = extrapolation.BOUNDARY,
             dims: tuple or None = None,
             stack_dim: str = 'gradient'):
    """
    Calculates the gradient of a scalar channel from finite differences.
    The gradient vectors are in reverse order, lowest dimension first.

    Args:
      grid: grid values
      dims: optional) sequence of dimension names
      dx: physical distance between grid points (default 1)
      difference: type of difference, one of ('forward', 'backward', 'central') (default 'forward')
      padding: tensor padding mode
      stack_dim: name of the new vector dimension listing the gradient w.r.t. the various axes
      grid: Tensor: 
      dx: float or int:  (Default value = 1)
      difference: str:  (Default value = 'central')
      padding: Extrapolation or None:  (Default value = extrapolation.BOUNDARY)
      dims: tuple or None:  (Default value = None)
      stack_dim: str:  (Default value = 'gradient')

    Returns:
      tensor of shape (batch_size, spatial_dimensions..., spatial rank)

    """
    grid = tensor(grid)
    if difference.lower() == 'central':
        left, right = shift(grid, (-1, 1), dims, padding, stack_dim=stack_dim)
        return (right - left) / (dx * 2)
    elif difference.lower() == 'forward':
        left, right = shift(grid, (0, 1), dims, padding, stack_dim=stack_dim)
        return (right - left) / dx
    elif difference.lower() == 'backward':
        left, right = shift(grid, (-1, 0), dims, padding, stack_dim=stack_dim)
        return (right - left) / dx
    else:
        raise ValueError('Invalid difference type: {}. Can be CENTRAL or FORWARD'.format(difference))


# Laplace

def laplace(x: Tensor,
            dx: Tensor or float = 1,
            padding: Extrapolation = extrapolation.BOUNDARY,
            dims: tuple or None = None):
    """
    Spatial Laplace operator as defined for scalar fields.
    If a vector field is passed, the laplace is computed component-wise.

    Args:
      x: n-dimensional field of shape (batch, spacial dimensions..., components)
      dx: scalar or 1d tensor
      padding: extrapolation
      dims: The second derivative along these dimensions is summed over
      x: Tensor: 
      dx: Tensor or float:  (Default value = 1)
      padding: Extrapolation:  (Default value = extrapolation.BOUNDARY)
      dims: tuple or None:  (Default value = None)

    Returns:
      tensor of same shape

    """
    if not isinstance(dx, (int, float)):
        dx = tensor(dx, names='_laplace')
    if isinstance(x, Extrapolation):
        return x.gradient()
    left, center, right = shift(tensor(x), (-1, 0, 1), dims, padding, stack_dim='_laplace')
    result = (left + right - 2 * center) / dx
    result = math.sum_(result, '_laplace')
    return result


def fourier_laplace(grid: Tensor,
                    dx: Tensor or Shape or float or list or tuple,
                    times: int = 1):
    """
    Applies the spatial laplace operator to the given tensor with periodic boundary conditions.
    
    *Note:* The results of `fourier_laplace` and `laplace` are close but not identical.
    
    This implementation computes the laplace operator in Fourier space.
    The result for periodic fields is exact, i.e. no numerical instabilities can occur, even for higher-order derivatives.

    Args:
      grid: tensor, assumed to have periodic boundary conditions
      dx: distance between grid points, tensor-like, scalar or vector
      times: number of times the laplace operator is applied. The computational cost is independent of this parameter.
      grid: Tensor: 
      dx: Tensor or Shape or float or list or tuple: 
      times: int:  (Default value = 1)

    Returns:
      tensor of same shape as `tensor`

    """
    frequencies = math.fft(math.to_complex(grid))
    k_squared = math.sum_(math.fftfreq(grid.shape) ** 2, 'vector')
    fft_laplace = -(2 * np.pi)**2 * k_squared
    result = math.real(math.ifft(frequencies * fft_laplace ** times))
    return math.cast(result / tensor(dx) ** 2, grid.dtype)


def fourier_poisson(grid: Tensor,
                    dx: Tensor or Shape or float or list or tuple,
                    times: int = 1):
    """
    Inverse operation to `fourier_laplace`.

    Args:
      grid: Tensor: 
      dx: Tensor or Shape or float or list or tuple: 
      times: int:  (Default value = 1)

    Returns:

    """
    frequencies = math.fft(math.to_complex(grid))
    k_squared = math.sum_(math.fftfreq(grid.shape) ** 2, 'vector')
    fft_laplace = -(2 * np.pi)**2 * k_squared
    # fft_laplace.tensor[(0,) * math.ndims(k_squared)] = math.inf  # assume NumPy array to edit
    result = math.real(math.ifft(math.divide_no_nan(frequencies, math.to_complex(fft_laplace ** times))))
    return math.cast(result * tensor(dx) ** 2, grid.dtype)


# Downsample / Upsample

def downsample2x(grid: Tensor,
                 padding: Extrapolation = extrapolation.BOUNDARY,
                 dims: tuple or None = None) -> Tensor:
    """
    Resamples a regular grid to half the number of spatial sample points per dimension.
    The grid values at the new points are determined via linear interpolation.

    Args:
      grid: full size grid
      padding: grid extrapolation. Used to insert an additional value for odd spatial dims
      dims: dims along which down-sampling is applied. If None, down-sample along all spatial dims.
      grid: Tensor: 
      padding: Extrapolation:  (Default value = extrapolation.BOUNDARY)
      dims: tuple or None:  (Default value = None)

    Returns:
      half-size grid

    """
    dims = grid.shape.spatial.only(dims).names
    odd_dimensions = [dim for dim in dims if grid.shape.get_size(dim) % 2 != 0]
    grid = math.pad(grid, {dim: (0, 1) for dim in odd_dimensions}, padding)
    for dim in dims:
        grid = (grid[{dim: slice(1, None, 2)}] + grid[{dim: slice(0, None, 2)}]) / 2
    return grid


def upsample2x(grid: Tensor,
               padding: Extrapolation = extrapolation.BOUNDARY,
               dims: tuple or None = None) -> Tensor:
    """
    Resamples a regular grid to double the number of spatial sample points per dimension.
    The grid values at the new points are determined via linear interpolation.

    Args:
      grid: half-size grid
      padding: grid extrapolation
      dims: dims along which up-sampling is applied. If None, up-sample along all spatial dims.
      grid: Tensor: 
      padding: Extrapolation:  (Default value = extrapolation.BOUNDARY)
      dims: tuple or None:  (Default value = None)

    Returns:
      double-size grid

    """
    for i, dim in enumerate(grid.shape.spatial.only(dims).names):
        left, center, right = shift(grid, (-1, 0, 1), (dim,), padding, None)
        interp_left = 0.25 * left + 0.75 * center
        interp_right = 0.75 * center + 0.25 * right
        stacked = math.spatial_stack([interp_left, interp_right], '_interleave')
        grid = math.join_dimensions(stacked, (dim, '_interleave'), dim)
    return grid


def sample_subgrid(grid: Tensor, start: Tensor, size: Shape) -> Tensor:
    """
    Samples a sub-grid from `grid` with equal distance between sampling points.
    The values at the new sample points are determined via linear interpolation.

    Args:
      grid: full size grid to be resampled
      start: origin point of sub-grid within `grid`, measured in number of cells.
    Must have a single dimension called `vector`.
    Example: `start=(1, 0.5)` would slice off the first grid point in dim 1 and take the mean of neighbouring points in dim 2.
    The order of dims must be equal to `size` and `grid.shape.spatial`.
      size: resolution of the sub-grid. Must not be larger than the resolution of `grid`.
    The order of dims must be equal to `start` and `grid.shape.spatial`.
      grid: Tensor: 
      start: Tensor: 
      size: Shape: 

    Returns:
      sampled sub-grid

    """
    assert start.shape.names == ('vector',)
    assert grid.shape.spatial.names == size.names
    discard = {}
    for dim, d_start, d_size in zip(grid.shape.spatial.names, start, size):
        discard[dim] = slice(int(d_start), int(d_start) + d_size + (1 if d_start != 0 else 0))
    grid = grid[discard]
    upper_weight = start % 1
    lower_weight = 1 - upper_weight
    for i, dim in enumerate(grid.shape.spatial.names):
        if upper_weight[i] not in (0, 1):
            lower, upper = shift(grid, (0, 1), [dim], padding=None, stack_dim=None)
            grid = upper * upper_weight[i] + lower * lower_weight[i]
    return grid


# Poisson Brackets


def poisson_bracket(grid1, grid2):
    if all([grid1.rank == grid2.rank == 2,
            grid1.boundary == grid2.boundary == extrapolation.PERIODIC,
            len(set(list(grid1.dx) + list(grid2.dx))) == 1]):
        return _periodic_2d_arakawa_poisson_bracket(grid1.values, grid2.values, grid1.dx)
    else:
        raise NotImplementedError("\n".join[
            "Not implemented for:"
            f"ranks ({grid1.rank}, {grid2.rank}) != 2",
            f"boundary ({grid1.boundary}, {grid2.boundary}) != {extrapolation.PERIODIC}",
            f"dx uniform ({grid1.dx}, {grid2.dx})"
        ])


def _periodic_2d_arakawa_poisson_bracket(tensor1: Tensor, tensor2: Tensor, dx: float):
    """
    Solves the poisson bracket using the Arakawa Scheme [tensor1, tensor2]
    
    Only works in 2D, with equal spaced grids, and periodic boundary conditions

    Args:
      tensor1(Tensor): first field in the poisson bracket
      tensor2(Tensor): second field in the poisson bracket
      dx(float): Grid size (equal in x-y)
      tensor1: Tensor: 
      tensor2: Tensor: 
      dx: float: 

    Returns:

    """
    zeta = math.pad(value=tensor1, widths={'x': (1, 1), 'y': (1, 1)}, mode=extrapolation.PERIODIC)
    psi = math.pad(value=tensor2, widths={'x': (1, 1), 'y': (1, 1)}, mode=extrapolation.PERIODIC)
    return (zeta.x[2:].y[1:-1] * (psi.x[1:-1].y[2:] - psi.x[1:-1].y[0:-2] + psi.x[2:].y[2:] - psi.x[2:].y[0:-2])
            - zeta.x[0:-2].y[1:-1] * (psi.x[1:-1].y[2:] - psi.x[1:-1].y[0:-2] + psi.x[0:-2].y[2:] - psi.x[0:-2].y[0:-2])
            - zeta.x[1:-1].y[2:] * (psi.x[2:].y[1:-1] - psi.x[0:-2].y[1:-1] + psi.x[2:].y[2:] - psi.x[0:-2].y[2:])
            + zeta.x[1:-1].y[0:-2] * (psi.x[2:].y[1:-1] - psi.x[0:-2].y[1:-1] + psi.x[2:].y[0:-2] - psi.x[0:-2].y[0:-2])
            + zeta.x[2:].y[0:-2] * (psi.x[2:].y[1:-1] - psi.x[1:-1].y[0:-2])
            + zeta.x[2:].y[2:] * (psi.x[1:-1].y[2:] - psi.x[2:].y[1:-1])
            - zeta.x[0:-2].y[2:] * (psi.x[1:-1].y[2:] - psi.x[0:-2].y[1:-1])
            - zeta.x[0:-2].y[0:-2] * (psi.x[0:-2].y[1:-1] - psi.x[1:-1].y[0:-2])) / (12 * dx**2)



"""
Vectorized operations, tensors with named dimensions.

This package provides a common interface for tensor operations.
Is internally uses NumPy, TensorFlow or PyTorch.

Main classes: `Tensor`, `Shape`, `DType`, `Extrapolation`.

The provided operations are not implemented directly.
Instead, they delegate the actual computation to either NumPy, TensorFlow or PyTorch, depending on the configuration.
This allows the user to write simulation code once and have it run with various computation backends.

See the documentation at https://tum-pbs.github.io/PhiFlow/Math.html
"""

from .backend import precision, set_global_precision, get_precision, Solve, LinearSolve, DType, SCIPY_BACKEND

from .extrapolation import Extrapolation

from ._config import GLOBAL_AXIS_ORDER

from ._shape import Shape, spatial_shape, EMPTY_SHAPE, batch_shape, channel_shape, shape
from ._tensors import tensor, tensors, Tensor, TensorDim
from ._functions import (
    choose_backend_t as choose_backend, all_available,
    print_ as print,
    map_ as map,
    trace_function,
    zeros, ones, fftfreq, random_normal, random_uniform, meshgrid, linspace,  # creation operators (use default backend)
    zeros_like, ones_like,
    batch_stack, spatial_stack, channel_stack, unstack, concat,
    pad, spatial_pad,
    join_dimensions, split_dimension, flatten, expand, expand_batch, expand_spatial, expand_channel, transpose, # reshape operations
    divide_no_nan,
    where, nonzero,
    sum_ as sum, mean, std, prod, max_ as max, min_ as min, any_ as any, all_ as all,  # reduce
    dot, matmul, einsum,
    abs, sign,
    round_ as round, ceil, floor,
    maximum, minimum, clip,
    sqrt, exp, sin, cos,
    to_float, to_int, to_complex, imag, real,
    boolean_mask,
    isfinite,
    closest_grid_values, grid_sample, scatter,
    fft, ifft, conv,
    dtype, cast,
    tile,
    close, assert_close,
    solve, minimize,
    record_gradients, gradients, stop_gradient
)
from ._nd import (
    shift,
    spatial_sum, vec_abs, vec_squared, cross_product,
    normalize_to,
    l1_loss, l2_loss, l_n_loss, frequency_loss,
    gradient, laplace,
    fourier_laplace, fourier_poisson, abs_square,
    downsample2x, upsample2x, sample_subgrid,
    extrapolate_valid_values,
)

PI = 3.14159265358979323846
"""Value of π to double precision """
pi = PI

SCIPY_BACKEND = SCIPY_BACKEND  # to show up in pdoc
"""Default backend for NumPy arrays and SciPy objects."""

__all__ = [key for key in globals().keys() if not key.startswith('_')]



import numpy as np

from ._value import EditableValue
from typing import Callable


class Control(object):

    def __init__(self, model, attribute_name: str, editable_value: int or float):
        assert isinstance(editable_value, EditableValue)
        self.model = model
        self.attribute_name = attribute_name
        self.editable_value = editable_value

    @property
    def value(self):
        val = getattr(self.model, self.attribute_name)
        if isinstance(val, np.float32):
            return float(val)
        if isinstance(val, np.float64):
            return float(val)
        return val

    @value.setter
    def value(self, value):
        setattr(self.model, self.attribute_name, value)
        self.model.invalidate()

    @property
    def name(self):
        return self.editable_value.name

    @property
    def type(self):
        return self.editable_value.type

    @property
    def id(self):
        return self.attribute_name

    def __str__(self):
        return self.name + '_' + str(self.value)

    @property
    def range(self):
        return self.editable_value.minmax


class Action(object):

    def __init__(self, name: str, method: Callable, id: str):
        self.name = name
        self.method = method
        self.method_name = id

    @property
    def id(self):
        return self.method_name



from phi import math
from phi.math import Shape
from ._field import Field
from ..geom import Geometry


class ConstantField(Field):

    def __init__(self, value=1.0):
        self.value = math.tensor(value)

    @property
    def shape(self) -> Shape:
        return self.value.shape

    def _op1(self, operator) -> Field:
        return ConstantField(operator(self.value))

    def _op2(self, other, operator) -> Field:
        return ConstantField(operator(self.value, other))

    def sample_at(self, points, reduce_channels=()) -> math.Tensor:
        return self.value

    def unstack(self, dimension: str):
        return tuple(ConstantField(v) for v in self.value.unstack(dimension))

    def __repr__(self):
        return repr(self.value)



from phi import math
from ._geom import Geometry
from ..math._shape import shape_stack, Shape, BATCH_DIM


class GeometryStack(Geometry):

    def __init__(self, geometries, dim_name):
        self.geometries = tuple(geometries)
        self.stack_dim_name = dim_name
        self._shape = shape_stack(dim_name, BATCH_DIM, *[g.shape for g in geometries])

    def unstack(self, dimension):
        if dimension == self.stack_dim_name:
            return self.geometries
        else:
            return GeometryStack([g.unstack(dimension) for g in self.geometries], self.stack_dim_name)

    @property
    def center(self):
        centers = [g.center for g in self.geometries]
        return math.batch_stack(centers, self.stack_dim_name)

    @property
    def shape(self):
        return self._shape

    def lies_inside(self, location: math.Tensor):
        if self.stack_dim_name in location.shape:
            location = location.unstack(self.stack_dim_name)
        else:
            location = [location] * len(self.geometries)
        inside = [g.lies_inside(loc) for g, loc in zip(self.geometries, location)]
        return math.batch_stack(inside, self.stack_dim_name)

    def approximate_signed_distance(self, location: math.Tensor):
        raise NotImplementedError()

    def bounding_radius(self):
        radii = [g.bounding_radius() for g in self.geometries]
        return math.batch_stack(radii, self.stack_dim_name)

    def bounding_half_extent(self):
        values = [g.bounding_half_extent() for g in self.geometries]
        return math.batch_stack(values, self.stack_dim_name)

    def shifted(self, delta: math.Tensor):
        geometries = [g.shifted(delta) for g in self.geometries]
        return GeometryStack(geometries, self.stack_dim_name)

    def rotated(self, angle):
        geometries = [g.rotated(angle) for g in self.geometries]
        return GeometryStack(geometries, self.stack_dim_name)

    def __eq__(self, other):
        return isinstance(other, GeometryStack) \
               and self._shape == other.shape \
               and self.stack_dim_name == other.stack_dim_name \
               and self.geometries == other.geometries

    def __hash__(self):
        return hash(self.geometries)


def stack(*geometries: Geometry, dim: str):
    return GeometryStack(geometries, dim)



import torch

from phi import struct
from ._torch_backend import TORCH_BACKEND


def variable(initial_value, dtype=None, requires_grad=True, device=None, item_condition=struct.VARIABLES):
    def f(attr): return torch.tensor(attr.value, dtype=dtype if dtype is not None else TORCH_BACKEND.precision_dtype, requires_grad=requires_grad, device=device)
    return struct.map(f, initial_value, trace=True, item_condition=item_condition)


def torch_from_numpy(obj, item_condition=struct.DATA):
    return struct.map(torch.from_numpy, obj, item_condition=item_condition)


def torch_to_numpy(obj, item_condition=struct.ALL_ITEMS):
    def to_numpy(obj):
        if isinstance(obj, torch.Tensor):
            return obj.numpy()
        else:
            return obj
    return struct.map(to_numpy, obj, item_condition=item_condition)



# pylint: disable-msg = unused-import
"""
*Compatibility import for PhiFlow 1 projects:* `from phi.flow_1 import *`
"""

from .flow import *
from .physics._boundaries import Material, Obstacle, OPEN, CLOSED, NO_STICK, SLIPPERY, NO_SLIP, STICKY, PERIODIC, Domain, GeometryMovement
from .physics._effect import FieldEffect, GROW, ADD, FIX, Inflow, Accelerator, ConstantVelocity, HeatSource, ColdSource, Fan, Gravity
from .physics._world import World, world, obstacle_mask
from .physics._fluid_legacy import Fluid, IncompressibleFlow
from .physics.diffuse import explicit as diffuse
from phi.field._scene import write_sim_frame

physics_config = math.GLOBAL_AXIS_ORDER
physics_config.x_last()
box = Box
SampledField = PointCloud



"""
Open-source simulation toolkit built for optimization and machine learning applications.

Use one of the following imports:

* `from phi.flow import *`  for NumPy mode
* `from phi.tf.flow import *`  for TensorFlow mode
* `from phi.torch.flow import *` for PyTorch mode
* `from phi.jax.flow import *` for *experimental* Jax mode

Project homepage: https://github.com/tum-pbs/PhiFlow

Documentation overview: https://tum-pbs.github.io/PhiFlow

PyPI: https://pypi.org/project/phiflow/
"""

import os as _os


with open(_os.path.join(_os.path.dirname(__file__), 'VERSION'), 'r') as version_file:
    __version__ = version_file.read()


def verify():
    """
    Checks your configuration for potential problems and prints a summary.

    To run verify without importing `phi`, run the script `tests/verify.py` included in the source distribution.
    """
    import sys
    from ._troubleshoot import assert_minimal_config, troubleshoot
    try:
        assert_minimal_config()
    except AssertionError as fail_err:
        print("\n".join(fail_err.args), file=sys.stderr)
        return
    print(troubleshoot())



import warnings

import imageio
import numpy as np
import os
import inspect

import webglviewer
from phi.math import GLOBAL_AXIS_ORDER as physics_config
from phi.field import CenteredGrid, StaggeredGrid
from .dash_plotting import reduce_component


def load_sky(file, image_format=None, flatten=True, resolution=None, scale=1.0):
    # type: (str, str, bool, int, float) -> tuple
    assert image_format in ('cubemap', 'equiangular', None)
    imageio.plugins.freeimage.download()
    if not os.path.isabs(file):
        file = os.path.join(os.path.dirname(inspect.getfile(webglviewer)), file)
    image = imageio.imread(file)
    raw = np.array(image) * scale
    raw = np.concatenate((raw, np.ones((raw.shape[0], raw.shape[1], 1))), axis=-1)  # add alpha channel
    if image_format is None:
        image_format = detect_skybox_format(raw)
    if image_format == 'cubemap':
        images = split_cubemap(raw, resolution)
    else:
        images = equiangular_to_cubemap(raw, resolution)
    if flatten:
        images = [im.flatten() for im in images]
    return tuple(images)


def detect_skybox_format(raw):
    h, w = raw.shape[:2]
    if w == 2 * h:
        return 'equiangular'
    if h // 3 == w // 4:
        return 'cubemap'
    raise ValueError('Unknown skybox format. Shape=%s' % raw.shape)


def split_cubemap(raw, target_resolution):
    h, w = raw.shape[0] // 3, raw.shape[1] // 4
    assert w == h
    right = raw[h:2*h, 2*w:3*w]
    left = raw[h:2*h, :w]
    top = raw[:h, w:2*w]
    bottom = raw[2*h:, w:2*w]
    front = raw[h:2*h, w:2*w]
    back = raw[h:2*h, 3*w:]
    return right, left, top, bottom, front, back


def equiangular_to_cubemap(equiangular, resolution=None):
    shape = equiangular.shape
    h, w = shape[:2]

    if resolution is None:
        resolution = h // 2

    def lookup(theta, phi):
        """
        

        Args:
          theta: pi, pi]
          phi: 0, 2 pi]

        Returns:

        """
        flip = np.where((theta > np.pi/2) | (theta < -np.pi/2), -1, 1)
        phi *= flip
        theta = (theta * flip + np.pi/2) % np.pi - np.pi/2
        y = np.rint((-theta / np.pi + 0.5) * h).astype(np.int)
        y = np.clip(y, 0, h-1)
        x = np.rint((phi / 2 / np.pi + 0.5) * w).astype(np.int)
        x = x % w
        return equiangular[y, x, :]

    def face(theta0, phi0):
        x, y = np.meshgrid(*[np.linspace(-1, 1, resolution)] * 2)
        theta = np.arctan(y) + theta0
        phi = np.arctan(x) + phi0
        return theta, phi

    right = lookup(*face(0, np.pi/2))
    left = lookup(*face(0, -np.pi/2))
    top = lookup(*face(np.pi/2, 0))
    bottom = lookup(*face(-np.pi/2, 0))
    front = lookup(*face(0, 0))
    back = lookup(*face(0, np.pi))
    images = right, left, top, bottom, front, back
    return images


def default_sky():
    right = [.3, .5, .4]
    left = [.3, .4, .5]
    top = [.8, .8, 1]
    bottom = [.2, .15, .15]
    front = [.4, .5, .4]
    back = [.4, .4, .4]
    images = right, left, top, bottom, front, back
    images = [np.array(image + [1.0]) * 255 for image in images]
    return images


EMPTY_GRID = np.zeros([2, 2, 2], np.float32)


def webgl_prepare_data(field, settings):
    # type: (object, dict) -> np.ndarray
    if field is None:
        return EMPTY_GRID

    component = settings.get('component', 'length')
    batch = settings.get('batch', 0)

    data = None

    if isinstance(field, CenteredGrid):
        if field.spatial_rank == 1:
            return EMPTY_GRID
        else:
            data = field.values

    elif isinstance(field, StaggeredGrid):
        if field.spatial_rank == 1:
            return EMPTY_GRID
        if component == 'vec2' or component == 'length':
            data = field.at_centers().values
            component = 'length'
        else:
            data = field.unstack()[{'z': physics_config.z, 'y': physics_config.y, 'x': physics_config.x}[component]].data

    if data is None:
        return EMPTY_GRID
    if not isinstance(data, np.ndarray):
        warnings.warn("WebGL plotting failed. Expected NumPy array but found '%s'" % data)
        return EMPTY_GRID

    if data.ndim == 4:
        data = np.stack([data] * 2, axis=1 if not physics_config.is_x_first else -2)

    data = data[min(batch, data.shape[0] - 1), ...]
    data = reduce_component(data, component)
    if not physics_config.is_x_first:
        data = np.transpose(data, axes=(1, 0, 2))
    else:
        data = np.transpose(data, axes=(1, 2, 0))
    data = np.abs(data) * 2
    return data



"""
Functions to simulate diffusion processes on `phi.field.Field` objects.
"""
import warnings

from phi import math
from phi.field import ConstantField, Grid, Field, laplace, solve
from phi.field._field_math import FieldType, GridType


def explicit(field: FieldType,
             diffusivity: float or math.Tensor or Field,
             dt: float or math.Tensor,
             substeps: int = 1) -> FieldType:
    """
    Simulate a finite-time diffusion process of the form dF/dt = α · ΔF on a given `Field` FieldType with diffusion coefficient α.

    If `field` is periodic (set via `extrapolation='periodic'`), diffusion may be simulated in Fourier space.
    Otherwise, finite differencing is used to approximate the

    Args:
        field: CenteredGrid, StaggeredGrid or ConstantField
        diffusivity: Diffusion per time. `diffusion_amount = diffusivity * dt`
        dt: Time interval. `diffusion_amount = diffusivity * dt`
        substeps: number of iterations to use (Default value = 1)
        field: FieldType:

    Returns:
        Diffused field of same type as `field`.
    """
    amount = diffusivity * dt
    if isinstance(amount, Field):
        amount = amount.at(field)
    for i in range(substeps):
        field += amount / substeps * laplace(field)
    return field


def implicit(field: FieldType,
             diffusivity: float or math.Tensor or Field,
             dt: float or math.Tensor,
             order: int = 1,
             solve_params: math.Solve = math.LinearSolve(bake='sparse')) -> FieldType:
    """
    Diffusion by solving a linear system of equations.

    Args:
        order: Order of method, 1=first order. This translates to `substeps` for the explicit sharpening.
        field:
        diffusivity: Diffusion per time. `diffusion_amount = diffusivity * dt`
        dt: Time interval. `diffusion_amount = diffusivity * dt`
        solve_params:

    Returns:
        Diffused field of same type as `field`.
    """
    def sharpen(x):
        return explicit(x, diffusivity, -dt, substeps=order)

    converged, diffused, iterations = solve(sharpen, field, field, solve_params=solve_params)
    if math.all_available(converged):
        assert converged, f"Implicit diffusion solve did not converge after {iterations} iterations. Last estimate: {diffused.values}"
    return diffused


def fourier(field: GridType,
            diffusivity: float or math.Tensor,
            dt: float or math.Tensor) -> FieldType:
    """
    Exact diffusion of a periodic field in frequency space.

    For non-periodic fields or non-constant diffusivity, use another diffusion function such as `explicit()`.

    Args:
        field:
        diffusivity: Diffusion per time. `diffusion_amount = diffusivity * dt`
        dt: Time interval. `diffusion_amount = diffusivity * dt`

    Returns:
        Diffused field of same type as `field`.
    """
    if isinstance(field, ConstantField):
        return field
    assert isinstance(field, Grid), "Cannot diffuse field of type '%s'" % type(field)
    assert field.extrapolation == math.extrapolation.PERIODIC, "Fourier diffusion can only be applied to periodic fields."
    amount = diffusivity * dt
    k = math.fftfreq(field.resolution)
    k2 = math.vec_squared(k)
    fft_laplace = -(2 * math.PI) ** 2 * k2
    diffuse_kernel = math.exp(fft_laplace * amount)
    result_k = math.fft(field.values) * diffuse_kernel
    result_values = math.real(math.ifft(result_k))
    return field.with_(values=result_values)
